import numpy as np
import xarray as xr
from attrs import frozen
from scipy.linalg import lstsq
from scipy.optimize import least_squares

import primap2
from primap2.csg import StrategyUnableToProcess


@frozen
class GlobalLSStrategy:
    """Fill missing data by global least square matching.

    The NaNs in the first timeseries `ts(t)` are filled using harmonized data
    from the lower priority timeseries `fill_ts(t)`. For harmonization we use
    fill_ts(t)_h = fill_ts(t) * a + b,
    where fill_ts(t)_h is the harmonized dataset and a and b are determined by minimizing
    the least squares distance between ts(t) and fill_ts(t)_h.

    If the class is initialized with `allow_shift = True` the faster
    `scipy.linalg.lstsq` function is used and b can be arbitrary.
    For the case `allow_shift = False` (b = 0) `scipi.optimize.least_squares` is used.

    If there is no overlap in non-NaN data between ts(t) and fill_ts(t) a
    `StrategyUnableToProcess` error will be raised.

    If `allow_negative = False` and the harmonized time-series fill_ts(t)_h
    contains negative data a `StrategyUnableToProcess` error will be raised.
    """

    allow_shift: bool = True
    allow_negative: bool = False
    type = "globalLS"

    def _factor_mult(self, a, e, e_ref):
        return a * e - e_ref

    def _jac(self, a, e, e_ref):
        J = np.empty((e.size, 1))
        J[:, 0] = e
        return J

    def fill(
        self,
        *,
        ts: xr.DataArray,
        fill_ts: xr.DataArray,
        fill_ts_repr: str,
    ) -> tuple[xr.DataArray, list[primap2.ProcessingStepDescription]]:
        """Fill missing data by global least square matching.

        For a description of the algorithm, see the documentation of this class.

            Parameters
            ----------
            ts
                Base timeseries. Missing data (NaNs) in this timeseries will be filled.
                This function does not modify the data in ts.
            fill_ts
                Fill timeseries. Data from this timeseries will be used (possibly after
                modification) to fill missing data in the base timeseries.
                This function does not modify the data in fill_ts.
            fill_ts_repr
                String representation of fill_ts. Human-readable short representation of
                the fill_ts (e.g. the source).

            Returns
            -------
                filled_ts, descriptions. filled_ts contains the result, where missing
                data in ts is (partly) filled using scaled data from fill_ts.
                descriptions contains information about which years were affected and
                filled how.
        """

        filled_mask = ts.isnull() & ~fill_ts.isnull()
        time_filled = filled_mask["time"][filled_mask].to_numpy()

        if time_filled.any():
            # check if we have overlap. if not use substitution strategy
            # this might not be necessary because we initialize the LS algorithm with 1,
            # but better make it explicit
            overlap = ts.notnull() & fill_ts.notnull()
            if overlap.any():
                if self.allow_shift:
                    e = fill_ts[overlap.data].data
                    A = np.vstack((e, np.ones_like(e))).transpose()
                    e_ref = ts[overlap.data].data
                    x, res, rank, s = lstsq(A, e_ref)
                    fill_ts_harmo = fill_ts * x[0] + x[1]
                    if any(fill_ts_harmo < 0):
                        # use filling without shift
                        raise StrategyUnableToProcess(
                            reason="Negative data after harmonization excluded "
                            "by configuration"
                        )
                    else:
                        filled_ts = xr.core.ops.fillna(ts, fill_ts_harmo, join="exact")
                        descriptions = [
                            primap2.ProcessingStepDescription(
                                time=time_filled,
                                description=f"filled with least squares matched data from "
                                f"{fill_ts_repr}. a*x+b with a={x[0]:0.3f}, "
                                f"b={x[1]:0.3f}",
                                function=self.type,
                                source=fill_ts_repr,
                            )
                        ]
                else:
                    e = fill_ts[overlap.data].data
                    e_ref = ts[overlap.data].data
                    a0 = [1]  #  start with 1 as scaling factor
                    res = least_squares(
                        self._factor_mult, a0, jac=self._jac, args=(e, e_ref)
                    )

                    fill_ts_h = fill_ts * res["x"][0]
                    filled_ts = xr.core.ops.fillna(ts, fill_ts_h, join="exact")

                    descriptions = [
                        primap2.ProcessingStepDescription(
                            time=time_filled,
                            description="filled with least squares matched data from "
                            f"{fill_ts_repr}. Factor={res['x'][0]:0.3f}",
                            function=self.type,
                            source=fill_ts_repr,
                        )
                    ]
            else:
                raise StrategyUnableToProcess(
                    reason="No overlap between timeseries, can't match"
                )

        else:
            # if we don't have anything to fill we don't need to calculate anything
            filled_ts = ts
            descriptions = [
                primap2.ProcessingStepDescription(
                    time=time_filled,
                    description=f"no additional data in {fill_ts_repr}",
                    function=self.type,
                    source=fill_ts_repr,
                )
            ]

        return filled_ts, descriptions



import numpy as np
import pint
import xarray as xr


def allclose(a: xr.DataArray, b: xr.DataArray, *args, **kwargs) -> bool:
    """Like np.allclose, but converts a to b's units before comparing."""
    try:
        a = a.pint.to(b.pint.units)
    except pint.DimensionalityError:
        return False
    if a.dtype == float:  # need to use "allclose" to compare floats
        return np.allclose(a.pint.magnitude, b.pint.magnitude, *args, **kwargs)
    else:
        return (a.pint.magnitude == b.pint.magnitude).all()


def assert_equal(a: xr.DataArray, b: xr.DataArray, *args, **kwargs):
    """Asserts that contents are allclose(), and the name and attrs are also equal."""
    assert allclose(a, b, *args, **kwargs)
    assert a.attrs == b.attrs, (a.attrs, b.attrs)
    assert a.name == b.name, (a.name, b.name)


def assert_align(a: xr.DataArray, b: xr.DataArray) -> tuple[xr.DataArray, xr.DataArray]:
    """Asserts that a and b have the same shape and returns a and b with axes and
    dimensions aligned and sorted equally so that naive comparisons can be done."""
    assert set(a.dims) == set(b.dims), (a.dims, b.dims)
    aa, ba = xr.align(a, b, join="outer")
    aa = aa.transpose(*ba.dims)
    size_unchanged = sorted(aa.shape) == sorted(a.shape) and ba.shape == b.shape
    assert size_unchanged, (a.shape, b.shape)
    return aa, ba


def assert_aligned_equal(a: xr.DataArray, b: xr.DataArray, *args, **kwargs):
    """Assert that a and b are equal after alignment of their dimensions."""
    a, b = assert_align(a, b)
    assert_equal(a, b, *args, **kwargs)


def assert_ds_aligned_equal(a: xr.Dataset, b: xr.Dataset, *args, **kwargs):
    """Assert that a and b are equal after alignment of their dimensions."""
    assert set(a.keys()) == set(b.keys())
    for key in a.keys():
        assert_aligned_equal(a[key], b[key], *args, **kwargs)
    assert a.attrs == b.attrs, (a.attrs, b.attrs)



"""Summarization and descriptive statistics functions to get an overview of a data
set."""

import typing

import pandas as pd

from . import _accessor_base
from ._selection import alias_dims


class DataArrayOverviewAccessor(_accessor_base.BaseDataArrayAccessor):
    def to_df(self, name: str | None = None) -> pd.DataFrame | pd.Series:
        """Convert this array into an unstacked (i.e. non-tidy) pandas.DataFrame.

        Converting to an unstacked pandas.DataFrame is most useful for two-dimensional
        data because then there is no MultiIndex, making the result very easy to read.

        If you want a tidy dataframe, use xarray's da.to_dataframe() instead.

        Parameters
        ----------
        name: str
          Name to give to this array (required if unnamed).

        Returns
        -------
        df: pandas.DataFrame
        """
        if name is None:
            name = self._da.name
        pandas_obj = self._da.reset_coords(drop=True).to_dataframe(name)[name]
        pandas_obj.name = name
        if isinstance(pandas_obj, pd.DataFrame) or isinstance(
            pandas_obj.index, pd.MultiIndex
        ):
            return pandas_obj.unstack()
        else:  # Series without MultiIndex can't be unstacked, return them as-is
            return pandas_obj

    @alias_dims(["dims"])
    def coverage(self, *dims: typing.Hashable) -> pd.DataFrame | pd.Series:
        """Summarize how many data points exist for a dimension combination.

        For each combinations of values in the given dimensions, count the number of
        non-NaN data points in the array. The result is returned as an
        N-dimensional pandas DataFrame.

        If the array's dtype is ``bool``, count the number of True values instead. This
        makes it possible to easily apply preprocessing. For example, to count the
        number of valid time series use ``da.notnull().any("time").coverage(...)``.

        Parameters
        ----------
        *dims: str
            Names or aliases of the dimensions to be used for summarizing.
            You can specify any number of dimensions, but the readability
            of the result is best for one or two dimensions.

        Returns
        -------
        coverage: pandas.DataFrame or pandas.Series
            N-dimensional dataframe (series for N=1) summarizing the number of non-NaN
            data points for each combination of values in the given dimensions.
        """
        if not dims:
            raise ValueError("Specify at least one dimension.")
        da = self._da

        if da.name is None:
            name = "coverage"
        else:
            name = da.name

        if da.dtype != bool:
            da = da.notnull()

        return da.pr.sum(reduce_to_dim=dims).transpose(*dims).pr.to_df(name)


class DatasetOverviewAccessor(_accessor_base.BaseDatasetAccessor):
    def to_df(
        self,
        name: str | None = None,
    ) -> pd.DataFrame:
        """Convert this dataset into a pandas.DataFrame.

        It returns mostly the same as xarray's to_dataframe() method, but discards
        additional coordinates instead of including them in the output.

        Parameters
        ----------
        name: str, optional
          Name to give to the output columns.

        Returns
        -------
        df: pandas.DataFrame
        """
        df = self._ds.pr.remove_processing_info().reset_coords(drop=True).to_dataframe()
        if name is not None:
            df.columns.name = name
        return df

    @alias_dims(["dims"], additional_allowed_values=["entity"])
    def coverage(self, *dims: typing.Hashable) -> pd.DataFrame | pd.Series:
        """Summarize how many data points exist for a dimension combination.

        For each combinations of values in the given dimensions, count the number of
        non-NaN data points in the dataset. The result is returned as an
        N-dimensional pandas DataFrame.

        Only those data variables in the dataset are considered which are defined on
        all given dims, i.e. each dim is in ``ds[key].dims``.

        If the dataset only contains boolean arrays, count the number of True values
        instead. This makes it possible to easily apply preprocessing. For example,
        to count the number of valid time series use
        ``ds.notnull().any("time").coverage(...)``.

        Parameters
        ----------
        *dims: str
            Names or aliases of the dimensions to be used for summarizing.
            To use the name of the data variables (usually, the gases) as a coordinate,
            use "entity". You can specify any number of dimensions, but the readability
            of the result is best for one or two dimensions.

        Returns
        -------
        coverage: pandas.DataFrame or pandas.Series
            N-dimensional dataframe (series for N=1) summarizing the number of non-NaN
            data points for each combination of values in the given dimensions.
        """
        if not dims:
            raise ValueError("Specify at least one dimension.")

        ds = self._ds

        for dim in dims:
            if dim == "entity":
                continue
            ds = ds.drop_vars([x for x in ds if dim not in ds[x].dims])

        all_boolean = all(ds[var].dtype == bool for var in ds)
        if not all_boolean:  # Convert into boolean coverage array
            ds = ds.notnull()

        da = ds.pr.sum(reduce_to_dim=dims)
        if "entity" in dims:
            da = da.to_array("entity")

        return da.transpose(*dims).pr.to_df("coverage")



"""Tests for _data_reading.py"""

import datetime
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

import primap2
import primap2.pm2io as pm2io
import primap2.pm2io._conversion
from primap2.pm2io._data_reading import additional_coordinate_metadata

from .utils import assert_ds_aligned_equal

DATA_PATH = Path(__file__).parent / "data"


@pytest.mark.parametrize(
    "unit, entity, expected_attrs",
    [
        ("Mt", "CO2", {"units": "Mt", "entity": "CO2"}),
        (
            "Gg CO2",
            "KYOTOGHG (AR4GWP100)",
            {
                "units": "Gg CO2",
                "entity": "KYOTOGHG",
                "gwp_context": "AR4GWP100",
            },
        ),
        (
            "kg CO2",
            "CH4 (SARGWP100)",
            {
                "units": "kg CO2",
                "entity": "CH4",
                "gwp_context": "SARGWP100",
            },
        ),
    ],
)
def test_metadata_for_variable(unit, entity, expected_attrs):
    assert (
        pm2io._interchange_format.metadata_for_variable(unit, entity) == expected_attrs
    )


@pytest.mark.parametrize(
    "to_test_for_float, expected_result",
    [
        ("0.15", True),
        ("IE", False),
        ("NaN", True),
        (
            "25,000",
            False,
        ),  # be careful when reading data to process thousands seperators
        ("25,00", False),
        ("IE, NO", False),
    ],
)
def test_is_float(to_test_for_float, expected_result):
    assert pm2io._data_reading.is_float(to_test_for_float) == expected_result


@pytest.mark.parametrize(
    "code_to_test, expected_result",
    [
        ("IE", 0),
        ("NO", 0),
        ("-", 0),
        ("NO,NE", 0),
        ("NO, NE", 0),
        ("NE,NO", 0),
        ("NE, NO", 0),
        ("IE, NE, NO", 0),
        ("IE,NA,NO", 0),
        ("NA,IE,NO", 0),
        ("NO,NE,IE", 0),
    ],
)
def test_parse_code(code_to_test, expected_result):
    assert pm2io._data_reading.parse_code(code_to_test) == expected_result


@pytest.mark.parametrize(
    "code_to_test_nan, expected_result",
    [
        ("NE", np.nan),
        ("NE0", np.nan),
        ("C", np.nan),
        ("NaN", np.nan),
        ("nan", np.nan),
        ("NA, NE", np.nan),
        ("NA,NE", np.nan),
    ],
)
def test_parse_code_nan(code_to_test_nan, expected_result):
    assert np.isnan(pm2io._data_reading.parse_code(code_to_test_nan))


@pytest.mark.parametrize(
    "strs, user_na_conv, expected_result",
    [
        (
            ["IE", "IE,NA", "NA"],
            {},
            {"IE": 0, "IE,NA": 0, "NA": np.nan},
        ),
        (
            ["IE", "IE,NA", "NA"],
            {"NA": 0},
            {"IE": 0, "IE,NA": 0, "NA": 0},
        ),
    ],
)
def test_create_na_replacement_dict(strs, user_na_conv, expected_result):
    assert (
        pm2io._data_reading.create_str_replacement_dict(strs, user_na_conv)
        == expected_result
    )


def assert_attrs_equal(attrs_result, attrs_expected):
    assert attrs_result.keys() == attrs_expected.keys()
    assert attrs_result["attrs"] == attrs_expected["attrs"]
    assert attrs_result["time_format"] == attrs_expected["time_format"]
    assert attrs_result["dimensions"].keys() == attrs_expected["dimensions"].keys()
    for entity in attrs_result["dimensions"]:
        assert set(attrs_result["dimensions"][entity]) == set(
            attrs_expected["dimensions"][entity]
        )


@pytest.fixture
def coords_cols():
    return {
        "unit": "unit",
        "entity": "gas",
        "area": "country",
        "category": "category",
        "sec_cats__Class": "classification",
    }


@pytest.fixture
def add_coords_cols():
    return {"category_name": ["category_name", "category"]}


@pytest.fixture
def coords_defaults():
    return {
        "source": "TESTcsv2021",
        "sec_cats__Type": "fugitive",
        "scenario": "HISTORY",
    }


@pytest.fixture
def coords_terminologies():
    return {
        "area": "ISO3",
        "category": "IPCC2006",
        "sec_cats__Type": "type",
        "sec_cats__Class": "class",
        "scenario": "general",
    }


@pytest.fixture
def coords_value_mapping():
    return {
        "category": "PRIMAP1",
        "entity": "PRIMAP1",
        "unit": "PRIMAP1",
    }


@pytest.fixture
def coords_value_filling():
    return {
        "category": {  # col to fill
            "category_name": {  # col to fill from
                "Energy": "1",  # from value: to value
                "IPPU": "2",
            }
        }
    }


@pytest.fixture
def filter_keep():
    return {
        "f1": {"category": ["IPC0", "IPC2"]},
        "f2": {"classification": "TOTAL"},
    }


@pytest.fixture
def filter_remove():
    return {"f1": {"gas": "CH4"}, "f2": {"country": ["USA", "FRA"]}}


class TestReadWideCSVFile:
    def test_output(
        self,
        tmp_path,
        coords_cols,
        coords_defaults,
        coords_terminologies,
        coords_value_mapping,
        filter_keep,
        filter_remove,
    ):
        file_input = DATA_PATH / "test_csv_data_sec_cat.csv"
        file_expected = DATA_PATH / "test_read_wide_csv_file_output.csv"
        df_expected = pd.read_csv(file_expected, index_col=0)

        meta_data = {"references": "Just ask around."}

        df_result = pm2io.read_wide_csv_file_if(
            file_input,
            coords_cols=coords_cols,
            coords_defaults=coords_defaults,
            coords_terminologies=coords_terminologies,
            coords_value_mapping=coords_value_mapping,
            filter_keep=filter_keep,
            filter_remove=filter_remove,
            meta_data=meta_data,
        )
        attrs_result = df_result.attrs
        df_result.to_csv(tmp_path / "temp.csv")
        df_result = pd.read_csv(tmp_path / "temp.csv", index_col=0)
        pd.testing.assert_frame_equal(df_result, df_expected, check_column_type=False)

        attrs_expected = {
            "attrs": {
                "references": "Just ask around.",
                "sec_cats": ["Class (class)", "Type (type)"],
                "scen": "scenario (general)",
                "area": "area (ISO3)",
                "cat": "category (IPCC2006)",
            },
            "time_format": "%Y",
            "dimensions": {
                "*": [
                    "entity",
                    "source",
                    "area (ISO3)",
                    "Type (type)",
                    "unit",
                    "scenario (general)",
                    "Class (class)",
                    "category (IPCC2006)",
                ]
            },
        }

        assert_attrs_equal(attrs_result, attrs_expected)

    def test_no_sec_cats(
        self,
        tmp_path,
        coords_cols,
        coords_defaults,
        coords_terminologies,
        coords_value_mapping,
    ):
        file_input = DATA_PATH / "test_csv_data.csv"
        file_expected = DATA_PATH / "test_read_wide_csv_file_no_sec_cats.csv"
        df_expected = pd.read_csv(file_expected, index_col=0)

        del coords_cols["sec_cats__Class"]
        del coords_defaults["sec_cats__Type"]
        del coords_terminologies["sec_cats__Class"]
        del coords_terminologies["sec_cats__Type"]

        df_result = pm2io.read_wide_csv_file_if(
            file_input,
            coords_cols=coords_cols,
            coords_defaults=coords_defaults,
            coords_terminologies=coords_terminologies,
            coords_value_mapping=coords_value_mapping,
        )
        attrs_result = df_result.attrs
        df_result.to_csv(tmp_path / "temp.csv")
        df_result = pd.read_csv(tmp_path / "temp.csv", index_col=0)
        pd.testing.assert_frame_equal(df_result, df_expected, check_column_type=False)

        attrs_expected = {
            "attrs": {
                "scen": "scenario (general)",
                "area": "area (ISO3)",
                "cat": "category (IPCC2006)",
            },
            "time_format": "%Y",
            "dimensions": {
                "*": [
                    "entity",
                    "source",
                    "area (ISO3)",
                    "unit",
                    "scenario (general)",
                    "category (IPCC2006)",
                ]
            },
        }

        assert_attrs_equal(attrs_result, attrs_expected)

    def test_add_coords(
        self,
        tmp_path,
        coords_cols,
        add_coords_cols,
        coords_defaults,
        coords_terminologies,
        coords_value_mapping,
    ):
        file_input = DATA_PATH / "test_csv_data_category_name.csv"
        file_expected = DATA_PATH / "test_read_wide_csv_file_no_sec_cats_cat_name.csv"
        df_expected = pd.read_csv(file_expected, index_col=0)

        del coords_cols["sec_cats__Class"]
        del coords_defaults["sec_cats__Type"]
        del coords_terminologies["sec_cats__Class"]
        del coords_terminologies["sec_cats__Type"]

        df_result = pm2io.read_wide_csv_file_if(
            file_input,
            coords_cols=coords_cols,
            add_coords_cols=add_coords_cols,
            coords_defaults=coords_defaults,
            coords_terminologies=coords_terminologies,
            coords_value_mapping=coords_value_mapping,
        )
        attrs_result = df_result.attrs
        df_result.to_csv(tmp_path / "temp.csv")
        df_result = pd.read_csv(tmp_path / "temp.csv", index_col=0)
        pd.testing.assert_frame_equal(df_result, df_expected, check_column_type=False)

        attrs_expected = {
            "attrs": {
                "scen": "scenario (general)",
                "area": "area (ISO3)",
                "cat": "category (IPCC2006)",
            },
            "time_format": "%Y",
            "dimensions": {
                "*": [
                    "entity",
                    "source",
                    "area (ISO3)",
                    "unit",
                    "scenario (general)",
                    "category (IPCC2006)",
                ]
            },
            "additional_coordinates": {"category_name": "category (IPCC2006)"},
        }

        assert_attrs_equal(attrs_result, attrs_expected)

    def test_read_wide_fill_col(
        self,
        tmp_path,
        coords_cols,
        add_coords_cols,
        coords_defaults,
        coords_terminologies,
        coords_value_mapping,
        coords_value_filling,
    ):
        file_input = DATA_PATH / "test_csv_data_category_name_fill_cat_code.csv"
        file_expected = DATA_PATH / "test_read_wide_csv_file_no_sec_cats_cat_name.csv"
        df_expected = pd.read_csv(file_expected, index_col=0)

        del coords_cols["sec_cats__Class"]
        del coords_defaults["sec_cats__Type"]
        del coords_terminologies["sec_cats__Class"]
        del coords_terminologies["sec_cats__Type"]

        df_result = pm2io.read_wide_csv_file_if(
            file_input,
            coords_cols=coords_cols,
            add_coords_cols=add_coords_cols,
            coords_defaults=coords_defaults,
            coords_terminologies=coords_terminologies,
            coords_value_mapping=coords_value_mapping,
            coords_value_filling=coords_value_filling,
        )
        attrs_result = df_result.attrs
        df_result.to_csv(tmp_path / "temp.csv")
        df_result = pd.read_csv(tmp_path / "temp.csv", index_col=0)
        pd.testing.assert_frame_equal(df_result, df_expected, check_column_type=False)

        attrs_expected = {
            "attrs": {
                "scen": "scenario (general)",
                "area": "area (ISO3)",
                "cat": "category (IPCC2006)",
            },
            "time_format": "%Y",
            "dimensions": {
                "*": [
                    "entity",
                    "source",
                    "area (ISO3)",
                    "unit",
                    "scenario (general)",
                    "category (IPCC2006)",
                ]
            },
            "additional_coordinates": {"category_name": "category (IPCC2006)"},
        }

        assert_attrs_equal(attrs_result, attrs_expected)

    def test_entity_terminology(
        self,
        tmp_path,
        coords_cols,
        coords_defaults,
        coords_terminologies,
        coords_value_mapping,
    ):
        file_input = DATA_PATH / "test_csv_data.csv"
        file_expected = DATA_PATH / "test_read_wide_csv_file_no_sec_cats.csv"
        df_expected: pd.DataFrame = pd.read_csv(file_expected, index_col=0)
        df_expected = df_expected.rename(columns={"entity": "entity (PRIMAP1)"})

        del coords_cols["sec_cats__Class"]
        del coords_defaults["sec_cats__Type"]
        del coords_terminologies["sec_cats__Class"]
        del coords_terminologies["sec_cats__Type"]

        coords_terminologies["entity"] = "PRIMAP1"

        df_result = pm2io.read_wide_csv_file_if(
            file_input,
            coords_cols=coords_cols,
            coords_defaults=coords_defaults,
            coords_terminologies=coords_terminologies,
            coords_value_mapping=coords_value_mapping,
        )
        attrs_result = df_result.attrs
        df_result.to_csv(tmp_path / "temp.csv")
        df_result = pd.read_csv(tmp_path / "temp.csv", index_col=0)
        pd.testing.assert_frame_equal(df_result, df_expected, check_column_type=False)

        attrs_expected = {
            "attrs": {
                "scen": "scenario (general)",
                "area": "area (ISO3)",
                "cat": "category (IPCC2006)",
                "entity_terminology": "PRIMAP1",
            },
            "time_format": "%Y",
            "dimensions": {
                "*": [
                    "entity (PRIMAP1)",
                    "source",
                    "area (ISO3)",
                    "unit",
                    "scenario (general)",
                    "category (IPCC2006)",
                ]
            },
        }

        assert_attrs_equal(attrs_result, attrs_expected)

    def test_coords_value_mapping_dict(
        self,
        tmp_path,
        coords_cols,
        coords_defaults,
        coords_terminologies,
        filter_keep,
        filter_remove,
    ):
        file_input = DATA_PATH / "test_csv_data_sec_cat.csv"
        file_expected = DATA_PATH / "test_read_wide_csv_file_output.csv"
        df_expected = pd.read_csv(file_expected, index_col=0)

        coords_value_mapping = {
            "category": {"IPC1": "1", "IPC2": "2", "IPC3": "3", "IPC0": "0"},
            "entity": {"KYOTOGHG": "KYOTOGHG (SARGWP100)"},
            "unit": "PRIMAP1",
        }

        df_result = pm2io.read_wide_csv_file_if(
            file_input,
            coords_cols=coords_cols,
            coords_defaults=coords_defaults,
            coords_terminologies=coords_terminologies,
            coords_value_mapping=coords_value_mapping,
            filter_keep=filter_keep,
            filter_remove=filter_remove,
        )
        df_result.to_csv(tmp_path / "temp.csv")
        df_result = pd.read_csv(tmp_path / "temp.csv", index_col=0)
        pd.testing.assert_frame_equal(df_result, df_expected, check_column_type=False)

    def test_entity_default(
        self,
        tmp_path,
        coords_cols,
        coords_defaults,
        coords_terminologies,
        coords_value_mapping,
        filter_keep,
        filter_remove,
    ):
        file_input = DATA_PATH / "test_csv_data_sec_cat.csv"
        file_expected = DATA_PATH / "test_read_wide_csv_file_output_entity_def.csv"
        df_expected = pd.read_csv(file_expected, index_col=0)

        del coords_cols["entity"]
        del coords_value_mapping["entity"]
        coords_defaults["entity"] = "CO2"
        del filter_remove["f1"]

        df_result = pm2io.read_wide_csv_file_if(
            file_input,
            coords_cols=coords_cols,
            coords_defaults=coords_defaults,
            coords_terminologies=coords_terminologies,
            coords_value_mapping=coords_value_mapping,
            filter_keep=filter_keep,
            filter_remove=filter_remove,
        )
        df_result.to_csv(tmp_path / "temp.csv")
        df_result = pd.read_csv(tmp_path / "temp.csv", index_col=0)
        pd.testing.assert_frame_equal(df_result, df_expected, check_column_type=False)

    def test_unit_default(
        self,
        tmp_path,
        coords_cols,
        coords_defaults,
        coords_terminologies,
        coords_value_mapping,
        filter_keep,
        filter_remove,
    ):
        file_input = DATA_PATH / "test_csv_data_sec_cat.csv"
        file_expected = DATA_PATH / "test_read_wide_csv_file_output_unit_def.csv"
        df_expected = pd.read_csv(file_expected, index_col=0)

        del coords_cols["unit"]
        coords_defaults["unit"] = "Gg"
        filter_remove["f1"] = {"gas": "KYOTOGHG"}

        df_result = pm2io.read_wide_csv_file_if(
            file_input,
            coords_cols=coords_cols,
            coords_defaults=coords_defaults,
            coords_terminologies=coords_terminologies,
            coords_value_mapping=coords_value_mapping,
            filter_keep=filter_keep,
            filter_remove=filter_remove,
        )
        df_result.to_csv(tmp_path / "test.csv")
        df_result = pd.read_csv(tmp_path / "test.csv", index_col=0)
        pd.testing.assert_frame_equal(df_result, df_expected, check_column_type=False)

    def test_unit_harmonization(
        self,
        tmp_path,
        coords_cols,
        coords_defaults,
        coords_terminologies,
        coords_value_mapping,
    ):
        file_input = DATA_PATH / "test_csv_data_unit_harmonization.csv"
        file_expected = DATA_PATH / "test_read_wide_csv_file_output_unit_harm.csv"
        df_expected = pd.read_csv(file_expected, index_col=0)

        del coords_cols["sec_cats__Class"]
        del coords_defaults["sec_cats__Type"]
        del coords_terminologies["sec_cats__Class"]
        del coords_terminologies["sec_cats__Type"]

        df_result = pm2io.read_wide_csv_file_if(
            file_input,
            coords_cols=coords_cols,
            coords_defaults=coords_defaults,
            coords_terminologies=coords_terminologies,
            coords_value_mapping=coords_value_mapping,
        )
        df_result.to_csv(tmp_path / "test.csv")
        df_result = pd.read_csv(tmp_path / "test.csv", index_col=0)
        pd.testing.assert_frame_equal(df_result, df_expected, check_column_type=False)

    def test_function_mapping(
        self,
        tmp_path,
        coords_cols,
        coords_defaults,
        coords_terminologies,
        coords_value_mapping,
        filter_keep,
        filter_remove,
    ):
        file_input = DATA_PATH / "test_csv_data_sec_cat.csv"
        file_expected = DATA_PATH / "test_read_wide_csv_file_output_unit_def.csv"
        df_expected = pd.read_csv(file_expected, index_col=0)

        del coords_cols["unit"]
        coords_defaults["unit"] = "Gg"
        coords_value_mapping["category"] = (
            pm2io._conversion.convert_ipcc_code_primap_to_primap2
        )
        filter_remove["f1"] = {"gas": "KYOTOGHG"}

        df_result = pm2io.read_wide_csv_file_if(
            file_input,
            coords_cols=coords_cols,
            coords_defaults=coords_defaults,
            coords_terminologies=coords_terminologies,
            coords_value_mapping=coords_value_mapping,
            filter_keep=filter_keep,
            filter_remove=filter_remove,
        )
        df_result.to_csv(tmp_path / "test.csv")
        df_result = pd.read_csv(tmp_path / "test.csv", index_col=0)
        pd.testing.assert_frame_equal(df_result, df_expected, check_column_type=False)

    def test_col_missing(
        self,
        coords_cols,
        coords_defaults,
        coords_terminologies,
        coords_value_mapping,
        filter_keep,
        filter_remove,
    ):
        file_input = DATA_PATH / "test_csv_data_sec_cat.csv"

        coords_cols["sec_cats__Class"] = "class"

        with pytest.raises(ValueError, match="Columns {'class'} not found in CSV."):
            pm2io.read_wide_csv_file_if(
                file_input,
                coords_cols=coords_cols,
                coords_defaults=coords_defaults,
                coords_terminologies=coords_terminologies,
                coords_value_mapping=coords_value_mapping,
                filter_keep=filter_keep,
                filter_remove=filter_remove,
            )

    def test_unknown_mapping(
        self, coords_cols, coords_defaults, coords_terminologies, coords_value_mapping
    ):
        file_input = DATA_PATH / "test_csv_data_sec_cat.csv"

        coords_value_mapping["category"] = "non-existing"

        with pytest.raises(
            ValueError,
            match="Unknown metadata mapping 'non-existing' for column 'category'.",
        ):
            pm2io.read_wide_csv_file_if(
                file_input,
                coords_cols=coords_cols,
                coords_defaults=coords_defaults,
                coords_terminologies=coords_terminologies,
                coords_value_mapping=coords_value_mapping,
            )

    def test_overlapping_specification(
        self, coords_cols, coords_defaults, coords_terminologies, coords_value_mapping
    ):
        file_input = DATA_PATH / "test_csv_data_sec_cat.csv"

        coords_defaults["entity"] = "CO2"

        with pytest.raises(
            ValueError, match="{'entity'} given in coords_cols and coords_defaults."
        ):
            pm2io.read_wide_csv_file_if(
                file_input,
                coords_cols=coords_cols,
                coords_defaults=coords_defaults,
                coords_terminologies=coords_terminologies,
                coords_value_mapping=coords_value_mapping,
            )

    def test_overlapping_specification_add_coords(
        self, coords_cols, coords_defaults, coords_terminologies, coords_value_mapping
    ):
        file_input = DATA_PATH / "test_csv_data_sec_cat.csv"

        add_coords_cols = {"test": ["gas", "category"]}

        with pytest.raises(
            ValueError, match="{'gas'} given in coords_cols and add_coords_cols."
        ):
            pm2io.read_wide_csv_file_if(
                file_input,
                coords_cols=coords_cols,
                add_coords_cols=add_coords_cols,
                coords_defaults=coords_defaults,
                coords_terminologies=coords_terminologies,
                coords_value_mapping=coords_value_mapping,
            )

    def test_no_unit(
        self, coords_cols, coords_defaults, coords_terminologies, coords_value_mapping
    ):
        file_input = DATA_PATH / "test_csv_data_sec_cat.csv"

        del coords_cols["unit"]

        with pytest.raises(ValueError, match="Mandatory dimension 'unit' not defined."):
            pm2io.read_wide_csv_file_if(
                file_input,
                coords_cols=coords_cols,
                coords_defaults=coords_defaults,
                coords_terminologies=coords_terminologies,
                coords_value_mapping=coords_value_mapping,
            )

    def test_no_entity(
        self, coords_cols, coords_defaults, coords_terminologies, coords_value_mapping
    ):
        file_input = DATA_PATH / "test_csv_data_sec_cat.csv"

        del coords_cols["entity"]

        with pytest.raises(
            ValueError, match="Mandatory dimension 'entity' not defined."
        ):
            pm2io.read_wide_csv_file_if(
                file_input,
                coords_cols=coords_cols,
                coords_defaults=coords_defaults,
                coords_terminologies=coords_terminologies,
                coords_value_mapping=coords_value_mapping,
            )

    def test_unknown_category_mapping(
        self, coords_cols, coords_defaults, coords_terminologies, coords_value_mapping
    ):
        file_input = DATA_PATH / "test_csv_data_sec_cat.csv"

        coords_value_mapping["category"] = "TESTTEST"

        with pytest.raises(
            ValueError,
            match="Unknown metadata mapping 'TESTTEST' for column 'category'.",
        ):
            pm2io.read_wide_csv_file_if(
                file_input,
                coords_cols=coords_cols,
                coords_defaults=coords_defaults,
                coords_terminologies=coords_terminologies,
                coords_value_mapping=coords_value_mapping,
            )

    def test_unknown_entity_mapping(
        self, coords_cols, coords_defaults, coords_terminologies, coords_value_mapping
    ):
        file_input = DATA_PATH / "test_csv_data_sec_cat.csv"

        coords_value_mapping["entity"] = "TESTTEST"

        with pytest.raises(
            ValueError, match="Unknown metadata mapping 'TESTTEST' for column 'entity'."
        ):
            pm2io.read_wide_csv_file_if(
                file_input,
                coords_cols=coords_cols,
                coords_defaults=coords_defaults,
                coords_terminologies=coords_terminologies,
                coords_value_mapping=coords_value_mapping,
            )

    def test_unknown_coordinate(
        self, coords_cols, coords_defaults, coords_terminologies, coords_value_mapping
    ):
        file_input = DATA_PATH / "test_csv_data_sec_cat.csv"

        coords_defaults["citation"] = "this should go to attrs"

        with pytest.raises(
            ValueError,
            match="'citation' given in coords_defaults is unknown - prefix with "
            "'sec_cats__' to add a secondary category.",
        ):
            pm2io.read_wide_csv_file_if(
                file_input,
                coords_cols=coords_cols,
                coords_defaults=coords_defaults,
                coords_terminologies=coords_terminologies,
                coords_value_mapping=coords_value_mapping,
            )

    def test_unprocessed_strs(
        self,
        coords_cols,
        coords_defaults,
        coords_terminologies,
        coords_value_mapping,
    ):
        file_input = DATA_PATH / "test_csv_data_sec_cat_strings.csv"

        with pytest.raises(ValueError, match="String values"):
            pm2io.read_wide_csv_file_if(
                file_input,
                coords_cols=coords_cols,
                coords_defaults=coords_defaults,
                coords_terminologies=coords_terminologies,
                coords_value_mapping=coords_value_mapping,
                filter_keep={},
                filter_remove={},
                convert_str=False,
            )


class TestInterchangeFormat:
    def test_from(self):
        file_input = DATA_PATH / "test_read_wide_csv_file_output.csv"
        file_expected = DATA_PATH / "test_from_interchange_format_output.nc"
        ds_expected = primap2.open_dataset(file_expected)
        df_input = pd.read_csv(file_input, index_col=0)
        dims = [
            "area (ISO3)",
            "category (IPCC2006)",
            "scenario (general)",
            "Class (class)",
            "Type (type)",
            "unit",
            "entity",
            "source",
        ]
        attrs = {
            "attrs": {
                "area": "area (ISO3)",
                "cat": "category (IPCC2006)",
                "scen": "scenario (general)",
                "sec_cats": ["Class (class)", "Type (type)"],
            },
            "time_format": "%Y",
            "dimensions": {"*": dims},
        }
        ds_result = pm2io.from_interchange_format(df_input, attrs)
        assert_ds_aligned_equal(ds_result, ds_expected, equal_nan=True)

    def test_from_too_large(self, caplog):
        df = pd.DataFrame(
            {
                "a": np.arange(10),
                "b": np.arange(10),
                "c": np.arange(10),
                "entity": ["CO2"] * 10,
                "unit": ["Gg"] * 10,
                "2001": np.arange(10),
            }
        )
        df.attrs = {
            "attrs": {},
            "dimensions": {"CO2": ["a", "b", "c"]},
            "time_format": "%Y",
        }
        # projected array size should be 1000 > 100
        with pytest.raises(ValueError, match="Resulting array too large"):
            pm2io.from_interchange_format(df, max_array_size=100)

        assert "ERROR" in caplog.text
        assert (
            "Set with 1 entities and a total of 3 dimensions will have a size of 1,000"
            in caplog.text
        )

    def test_from_add_coord_non_unique(self, caplog):
        df = pd.DataFrame(
            {
                "a": np.arange(3),
                "b": np.arange(3),
                "c": np.arange(3),
                "entity": ["CO2"] * 3,
                "entity_name": ["Carbondioxide", "Carbondioxide", "Methane"],
                "unit": ["Gg"] * 3,
                "2001": np.arange(3),
            }
        )
        df.attrs = {
            "attrs": {},
            "dimensions": {"CO2": ["a", "b", "c"]},
            "time_format": "%Y",
            "additional_coordinates": {"entity_name": "entity"},
        }

        with pytest.raises(
            ValueError,
            match="Different secondary coordinate values "
            "for given first coordinate value for "
            "entity_name.",
        ):
            pm2io.from_interchange_format(df)

        assert "ERROR" in caplog.text
        assert (
            "Different secondary coordinate values for given first coordinate "
            "value for entity_name." in caplog.text
        )

    def test_roundtrip(self, tmp_path):
        file_input = DATA_PATH / "test_read_wide_csv_file_output.csv"
        file_temp = tmp_path / "test_interchange_format"
        data = pd.read_csv(file_input, index_col=0, dtype=object)
        attrs = {
            "attrs": {
                "area": "area (ISO3)",
                "cat": "category (IPCC2006)",
                "scen": "scenario (general)",
                "sec_cats": ["Class (class)", "Type (type)"],
            },
            "time_format": "%Y",
            "dimensions": {"CO2": ["area (ISO3)"]},
        }
        pm2io.write_interchange_format(file_temp, data, attrs)
        read_data = pm2io.read_interchange_format(file_temp)
        read_attrs = read_data.attrs
        assert read_attrs == attrs
        pd.testing.assert_frame_equal(data, read_data)


class TestLong:
    def test_compare_wide(
        self,
        coords_cols,
        coords_defaults,
        coords_terminologies,
        coords_value_mapping,
    ):
        file_input_wide = DATA_PATH / "test_csv_data.csv"
        file_input_long = DATA_PATH / "long.csv"

        del coords_cols["sec_cats__Class"]
        del coords_defaults["sec_cats__Type"]
        del coords_terminologies["sec_cats__Class"]
        del coords_terminologies["sec_cats__Type"]

        meta_data = {"references": "Just ask around"}

        df_result_wide = pm2io.read_wide_csv_file_if(
            file_input_wide,
            coords_cols=coords_cols,
            coords_defaults=coords_defaults,
            coords_terminologies=coords_terminologies,
            coords_value_mapping=coords_value_mapping,
            meta_data=meta_data,
        )

        coords_cols["time"] = "year"
        coords_cols["data"] = "emissions"
        df_result_long = pm2io.read_long_csv_file_if(
            file_input_long,
            coords_cols=coords_cols,
            coords_defaults=coords_defaults,
            coords_terminologies=coords_terminologies,
            coords_value_mapping=coords_value_mapping,
            meta_data=meta_data,
            time_format="%Y",
        )

        pd.testing.assert_frame_equal(df_result_wide, df_result_long)
        assert df_result_wide.attrs == df_result_long.attrs

    def test_compare_wide_entity_terminology(
        self,
        coords_cols,
        coords_defaults,
        coords_terminologies,
        coords_value_mapping,
    ):
        file_input_wide = DATA_PATH / "test_csv_data.csv"
        file_input_long = DATA_PATH / "long.csv"

        del coords_cols["sec_cats__Class"]
        del coords_defaults["sec_cats__Type"]
        del coords_terminologies["sec_cats__Class"]
        del coords_terminologies["sec_cats__Type"]

        coords_terminologies["entity"] = "PRIMAP1"

        df_result_wide = pm2io.read_wide_csv_file_if(
            file_input_wide,
            coords_cols=coords_cols,
            coords_defaults=coords_defaults,
            coords_terminologies=coords_terminologies,
            coords_value_mapping=coords_value_mapping,
        )

        coords_cols["time"] = "year"
        coords_cols["data"] = "emissions"
        df_result_long = pm2io.read_long_csv_file_if(
            file_input_long,
            coords_cols=coords_cols,
            coords_defaults=coords_defaults,
            coords_terminologies=coords_terminologies,
            coords_value_mapping=coords_value_mapping,
            time_format="%Y",
        )

        pd.testing.assert_frame_equal(df_result_wide, df_result_long)
        assert df_result_wide.attrs == df_result_long.attrs

    def test_compare_wide_add_cols(
        self,
        coords_cols,
        add_coords_cols,
        coords_defaults,
        coords_terminologies,
        coords_value_mapping,
    ):
        file_input_wide = DATA_PATH / "test_csv_data_category_name.csv"
        file_input_long = DATA_PATH / "test_csv_data_category_name_long.csv"

        del coords_cols["sec_cats__Class"]
        del coords_defaults["sec_cats__Type"]
        del coords_terminologies["sec_cats__Class"]
        del coords_terminologies["sec_cats__Type"]

        df_result_wide = pm2io.read_wide_csv_file_if(
            file_input_wide,
            coords_cols=coords_cols,
            add_coords_cols=add_coords_cols,
            coords_defaults=coords_defaults,
            coords_terminologies=coords_terminologies,
            coords_value_mapping=coords_value_mapping,
        )

        coords_cols["time"] = "year"
        coords_cols["data"] = "emissions"
        df_result_long = pm2io.read_long_csv_file_if(
            file_input_long,
            coords_cols=coords_cols,
            add_coords_cols=add_coords_cols,
            coords_defaults=coords_defaults,
            coords_terminologies=coords_terminologies,
            coords_value_mapping=coords_value_mapping,
            time_format="%Y",
        )

        pd.testing.assert_frame_equal(df_result_wide, df_result_long)
        assert df_result_wide.attrs == df_result_long.attrs

    def test_no_data_specified(
        self, coords_cols, coords_defaults, coords_terminologies, coords_value_mapping
    ):
        file_input_long = DATA_PATH / "long.csv"

        del coords_cols["sec_cats__Class"]
        del coords_defaults["sec_cats__Type"]
        del coords_terminologies["sec_cats__Class"]
        del coords_terminologies["sec_cats__Type"]

        coords_cols["time"] = "year"

        with pytest.raises(
            ValueError,
            match="No data column in the CSV specified in coords_cols, so nothing to"
            " read.",
        ):
            pm2io.read_long_csv_file_if(
                file_input_long,
                coords_cols=coords_cols,
                coords_defaults=coords_defaults,
                coords_terminologies=coords_terminologies,
                coords_value_mapping=coords_value_mapping,
                time_format="%Y",
            )

    def test_no_time(
        self, coords_cols, coords_defaults, coords_terminologies, coords_value_mapping
    ):
        file_input_long = DATA_PATH / "long_no_time.csv"

        del coords_cols["sec_cats__Class"]
        del coords_defaults["sec_cats__Type"]
        del coords_terminologies["sec_cats__Class"]
        del coords_terminologies["sec_cats__Type"]

        coords_cols["data"] = "emissions"
        coords_defaults["time"] = datetime.datetime(2020, 1, 1)

        df = pm2io.read_long_csv_file_if(
            file_input_long,
            coords_cols=coords_cols,
            coords_defaults=coords_defaults,
            coords_terminologies=coords_terminologies,
            coords_value_mapping=coords_value_mapping,
            time_format="%Y",
        )

        df_expected = pd.DataFrame(
            [
                ("TESTcsv2021", "HISTORY", "AUS", "CO2", "Gg CO2 / yr", "1", 4.1),
                ("TESTcsv2021", "HISTORY", "ZAM", "CH4", "Mt CH4 / yr", "2", 7.0),
            ],
            columns=[
                "source",
                "scenario (general)",
                "area (ISO3)",
                "entity",
                "unit",
                "category (IPCC2006)",
                "2020",
            ],
        )

        pd.testing.assert_frame_equal(df, df_expected)


class TestAdditionalCoordinateMetadata:
    def test_error_coord_terminology(self):
        add_coords_cols = {"category_name": ["cat_name", "category"]}
        coords_cols = {"category": "cat_code"}
        coords_terminologies = {"category_name": "IPCC2006_name"}

        with pytest.raises(
            ValueError,
            match="Additional coordinate category_name has terminology definition. "
            "This is currently not supported by PRIMAP2.",
        ):
            additional_coordinate_metadata(
                coords_cols=coords_cols,
                add_coords_cols=add_coords_cols,
                coords_terminologies=coords_terminologies,
            )

    def test_error_coord_not_present(self):
        add_coords_cols = {"category_name": ["cat_name", "category"]}
        coords_cols = {"entity": "entity"}
        coords_terminologies = {}

        with pytest.raises(
            ValueError,
            match="Additional coordinate category_name refers to unknown "
            "coordinate category",
        ):
            additional_coordinate_metadata(
                coords_cols=coords_cols,
                add_coords_cols=add_coords_cols,
                coords_terminologies=coords_terminologies,
            )


class TestRegression:
    def test_read_published_data(self):
        actual = pm2io.from_interchange_format(
            pm2io.read_interchange_format(
                DATA_PATH / "Guetschow-et-al-2021-PRIMAP-crf96_2021-v1"
            )
        )
        expected = primap2.open_dataset(
            DATA_PATH / "Guetschow-et-al-2021-PRIMAP-crf96_2021-v1.nc"
        )
        assert_ds_aligned_equal(actual, expected, equal_nan=True)


# functions that still need individual testing
# dates_to_dimension(ds: xr.Dataset, time_format: str = "%Y") -> xr.DataArray:
# harmonize_units



"""Tests for _overview.py"""

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from primap2 import ureg


def test_to_df():
    data = np.array([[1, 2], [3, 4]], dtype=np.int64)
    a = ["a1", "a2"]
    b = ["b1", "b2"]
    da = xr.DataArray(data, coords=[("a", a), ("b", b)], name="name")
    actual = da.pr.to_df()

    expected = pd.DataFrame(data, index=a, columns=b)
    expected.index.name = "a"
    expected.columns.name = "b"

    pd.testing.assert_frame_equal(actual, expected)


def test_to_df_1d():
    data = np.array([1, 2], dtype=np.int64)
    a = ["a1", "a2"]
    da = xr.DataArray(data, coords=[("a", a)], name="name")
    actual = da.pr.to_df()

    expected = pd.Series(data, index=a, name="name")
    expected.index.name = "a"

    pd.testing.assert_series_equal(actual, expected)


def test_to_df_set():
    data = np.array([1, 2], dtype=np.int64)
    a = ["a1", "a2"]
    da = xr.DataArray(data, coords=[("a", a)], name="name")
    ds = xr.Dataset({"b": da})
    actual = ds.pr.to_df("name")

    expected = pd.DataFrame(data, index=a, columns=["b"])
    expected.index.name = "a"
    expected.columns.name = "name"

    pd.testing.assert_frame_equal(actual, expected)


def test_array_empty(empty_ds):
    with pytest.raises(ValueError, match="Specify at least one dimension"):
        empty_ds.pr.coverage()
    with pytest.raises(ValueError, match="Specify at least one dimension"):
        empty_ds["CO2"].pr.coverage()


def test_array_coverage(empty_ds):
    da = empty_ds["CO2"]
    da.pint.magnitude[:] = np.nan
    da.name = None

    da.pr.loc[{"time": "2001", "area": "COL"}] = 12.0 * ureg("Gg CO2 / year")
    da.pr.loc[{"time": "2002", "area": "COL"}] = 13.0 * ureg("Gg CO2 / year")

    expected = pd.DataFrame(
        index=da["area (ISO3)"].values,
        columns=da["time"].to_index(),
        data=np.zeros((len(da["area (ISO3)"]), len(da["time"])), dtype=np.int32),
    )
    expected.loc["COL", "2001"] = 1
    expected.loc["COL", "2002"] = 1
    expected.index.name = "area (ISO3)"
    expected.columns.name = "time"

    pd.testing.assert_frame_equal(
        expected.astype(np.int32), da.pr.coverage("area", "time").astype(np.int32)
    )
    pd.testing.assert_frame_equal(
        expected.T.astype(np.int32),
        da.pr.coverage("time", "area (ISO3)").astype(np.int32),
    )


def test_array_coverage_multidim(opulent_ds):
    da = opulent_ds["CO2"]

    da.pr.loc[{"product": "milk"}].pint.magnitude[:] = np.nan

    expected = pd.DataFrame(
        index=da.pr["animal"].values,
        columns=da.pr["product"].values,
        data=np.zeros((len(da.pr["animal"]), len(da.pr["product"])), dtype=np.int32),
    )
    expected[:] = np.prod(da.shape) // np.prod(expected.shape)
    expected.loc[:, "milk"] = 0
    expected.index.name = "animal (FAOSTAT)"
    expected.columns.name = "product (FAOSTAT)"

    pd.testing.assert_frame_equal(
        expected.astype(np.int32), da.pr.coverage("animal", "product").astype(np.int32)
    )
    pd.testing.assert_frame_equal(
        expected.T.astype(np.int32),
        da.pr.coverage("product", "animal").astype(np.int32),
    )


def test_array_coverage_error(opulent_ds):
    da = opulent_ds["CO2"]

    with pytest.raises(ValueError, match="Dimension 'non-existing' does not exist."):
        da.pr.coverage("animal", "non-existing")


def test_set_coverage(opulent_ds):
    ds = opulent_ds
    ds["CO2"].pr.loc[{"product": "milk"}].pint.magnitude[:] = np.nan

    expected = pd.DataFrame(
        index=ds.pr["product"].values,
        columns=ds.pr["animal"].values,
        data=np.zeros((len(ds.pr["product"]), len(ds.pr["animal"])), dtype=int),
    )
    expected[:] = np.prod(ds["CO2"].shape) // np.prod(expected.shape) * 4
    expected.loc["milk", :] = np.prod(ds["CO2"].shape) // np.prod(expected.shape) * 3
    expected.index.name = "product (FAOSTAT)"
    expected.columns.name = "animal (FAOSTAT)"
    expected.name = "coverage"

    pd.testing.assert_frame_equal(expected, ds.pr.coverage("product", "animal"))
    pd.testing.assert_frame_equal(expected.T, ds.pr.coverage("animal", "product"))


def test_set_coverage_entity(opulent_ds):
    ds = opulent_ds
    ds["CO2"].pr.loc[{"product": "milk"}].pint.magnitude[:] = np.nan

    expected = pd.DataFrame(
        index=list(ds.keys()),
        columns=ds.pr["area"].values,
        data=np.zeros((len(ds), len(ds.pr["area"].values)), dtype=int),
    )
    expected[:] = np.prod(ds["CO2"].shape)
    expected.loc["population", :] = np.prod(ds["population"].shape)
    expected.loc["CO2", :] = np.prod(ds["CO2"].shape) - np.prod(
        ds["CO2"].pr.loc[{"product": "milk"}].shape
    )
    expected = expected // len(ds.pr["area"].values)
    expected.name = "coverage"
    expected.index.name = "entity"
    expected.columns.name = "area (ISO3)"

    pd.testing.assert_frame_equal(expected, ds.pr.coverage("entity", "area"))


def test_set_coverage_boolean(opulent_ds):
    actual = opulent_ds.notnull().any("time").pr.coverage("entity", "area")
    expected = opulent_ds.pr.coverage("entity", "area") // len(opulent_ds["time"])

    pd.testing.assert_frame_equal(actual, expected)


def test_set_coverage_entity_other_dim_not_existing(opulent_ds):
    ds = opulent_ds

    ds["CO2"].pr.loc[{"product": "milk"}].pint.magnitude[:] = np.nan

    entites_expected = [x for x in ds.keys() if x != "population"]

    expected = pd.DataFrame(
        index=ds.pr["product"].values,
        columns=entites_expected,
        data=np.zeros((len(ds.pr["product"]), len(entites_expected)), dtype=int),
    )
    expected[:] = np.prod(ds["CO2"].shape) // len(ds.pr["product"])
    expected.loc["milk", "CO2"] = 0
    expected.index.name = "product (FAOSTAT)"
    expected.columns.name = "entity"

    pd.testing.assert_frame_equal(expected, ds.pr.coverage("product", "entity"))
    pd.testing.assert_frame_equal(expected.T, ds.pr.coverage("entity", "product"))


def test_set_coverage_error(opulent_ds):
    ds = opulent_ds["CO2"]

    with pytest.raises(ValueError, match="Dimension 'non-existing' does not exist."):
        ds.pr.coverage("animal", "non-existing")



"""Merge arrays and datasets with optional tolerances."""

import contextlib
from datetime import date

import numpy as np
import xarray as xr
from loguru import logger

from ._accessor_base import BaseDataArrayAccessor, BaseDatasetAccessor


def merge_with_tolerance_core(
    *,
    da_start: xr.DataArray,
    da_merge: xr.DataArray,
    tolerance: float = 0.01,
    error_on_discrepancy: bool = True,
) -> xr.DataArray:
    """
    Merge two DataArrays with a given tolerance for discrepancies in values
    present in both DataArrays.

    If values from the data to merge are already present in da_start they are
    treated as equal if the relative difference is below the tolerance threshold.
    The result will use the values present in da_start.

    If a merge using xr.merge is not possible the function identifies the conflicting
    values and checks if the relative difference is above the tolerance threshold.
    If it is, an error is raised (if error_on_discrepancy = True) or a warning is
    logged (if error_on_discrepancy = False).

    The function assumes dataarrays have been checked for identical coordinates
    etc. as it is called by DataArray.pr.merge() and Dataset.pr.merge().

    Parameters
    ----------
    da_start: xr.DataArray
        data to merge into
    da_merge: xr.DataArray
        data to merge into ds_start
    tolerance: float (optional), default = 0.01
        The tolerance to use when comparing data. Tolerance is relative to values in
        the calling Dataset. Thus by default a 1% deviation of values in da_merge
        from the calling Dataset is tolerated.
    error_on_discrepancy: (optional), default = True
        If true throw an exception if false a warning and return values from
        the calling object in cases of conflict.

    Returns
    -------
        xr.DataArray: DataArray with data from da_merge merged into da_start
    """
    with contextlib.suppress(xr.MergeError):
        da_result = xr.merge(
            [da_start, da_merge],
            compat="no_conflicts",
            join="outer",
        )
        # all done as no errors occurred and thus no duplicates were present
        # make sure we have a DataArray not a Dataset
        return da_result[da_start.name]
    # there are conflicts (overlapping coordinates) between da_start and da_merge

    # calculate the deviation between da_start and da_merge
    da_comp = abs(da_start - da_merge) / da_start
    da_error = da_comp.where(da_comp > tolerance, drop=True)

    if np.logical_not(da_error.isnull()).any():
        # there are differences larger than the tolerance
        log_message = generate_log_message(da_error=da_error, tolerance=tolerance)
        if error_on_discrepancy:
            logger.error(log_message)
            raise xr.MergeError(log_message)
        else:
            # log warning, continue with merging
            logger.warning(log_message)

    # all differences are within the tolerance or ignored, take the first
    # value everywhere
    return da_start.pr.combine_first(da_merge)


def generate_log_message(da_error: xr.DataArray, tolerance: float) -> str:
    """Generate a single large log message for all given errors.

    Strategy:
    * remove all length-1 dimensions and put them in description
    * convert the rest into a pandas dataframe for nice printing
    """
    scalar_dims = [dim for dim in da_error.dims if len(da_error[dim]) == 1]
    scalar_dims_format = []
    for dim in scalar_dims:
        if dim == "time":
            single_date = date.fromtimestamp(da_error[dim].item() / 1000000000)
            scalar_dims_format.append(f"{dim}={single_date.strftime('%Y')}")
        else:
            scalar_dims_format.append(f"{dim}={da_error[dim].item()}")
    scalar_dims_str = ", ".join(scalar_dims_format)
    da_error_dequ = da_error.squeeze(drop=True).pint.dequantify()
    if np.ndim(da_error_dequ.data) == 0:
        errors_str = str(da_error_dequ.data)
    else:
        errors_str = da_error_dequ.to_dataframe().dropna().to_string()

    return (
        f"pr.merge error: found discrepancies larger than tolerance "
        f"({tolerance * 100:.2f}%) for {scalar_dims_str}:\n"
        f"shown are relative discrepancies. ({da_error.name})\n" + errors_str
    )


def ensure_compatible_coords_dims(
    a: xr.Dataset | xr.DataArray, b: xr.Dataset | xr.DataArray
):
    """Check if coordinates and dimensions of both Datasets or DataArrays agree,
    raise exception otherwise."""
    if set(a.coords) != set(b.coords):
        logger.error("pr.merge error: coords of objects to merge must agree")
        raise ValueError("pr.merge error: coords of objects to merge must agree")
    if set(a.dims) != set(b.dims):
        logger.error("pr.merge error: dims of objects to merge must agree")
        raise ValueError("pr.merge error: dims of objects to merge must agree")


class DataArrayMergeAccessor(BaseDataArrayAccessor):
    def merge(
        self,
        da_merge: xr.DataArray,
        tolerance: float = 0.01,
        error_on_discrepancy: bool = True,
    ) -> xr.DataArray:
        """
        Merge this data array with another using a given tolerance for
        discrepancies in values present in both DataArrays.

        If values from the data to merge are already
        present they are treated as equal if the relative
        difference is below the tolerance threshold.

        Parameters
        ----------
        da_merge: xr.DataArray
            data to merge to the calling object
        tolerance: float (optional), default = 0.01
            The tolerance to use when comparing data. Tolerance is relative to values in
            the calling Dataset. Thus by default a 1% deviation of values in da_merge
            from the calling Dataset is tolerated.
        error_on_discrepancy: (optional), default = True
            If true throw an exception if false a warning and return values from
            the calling object in cases of conflict.
        combine_attrs (optional), default = "drop_conflicts"
            Governs how to combine conflicting attrs. Is passed on to the xr merge
            functions.

        Returns
        -------
            xr.DataArray: DataArray with data from da_merge merged into the calling
            object
        """

        # check if coordinates and dimensions agree
        da_start = self._da
        ensure_compatible_coords_dims(da_start, da_merge)

        return merge_with_tolerance_core(
            da_start=da_start,
            da_merge=da_merge,
            tolerance=tolerance,
            error_on_discrepancy=error_on_discrepancy,
        )


class DatasetMergeAccessor(BaseDatasetAccessor):
    def merge(
        self,
        ds_merge: xr.Dataset,
        tolerance: float = 0.01,
        error_on_discrepancy: bool = True,
        combine_attrs: xr.core.types.CombineAttrsOptions = "drop_conflicts",
    ) -> xr.Dataset:
        """
        Merge two Datasets with a given tolerance for discrepancies in values
        present in both Datasets. If values from the data to merge are already
        present in the calling object they are treated as equal if the relative
        difference is below the tolerance threshold. The result will use the values
        of the calling object.

        Parameters
        ----------
        ds_merge: xr.Dataset
            data to merge to the calling object
        tolerance: float (optional), default = 0.01
            The tolerance to use when comparing data. Tolerance is relative to values in
            the calling Dataset. Thus, by default a 1% deviation of values in da_merge
            from the calling Dataset is tolerated.
        error_on_discrepancy: (optional), default = True
            If true throw an exception if false a warning and return values from
            the calling object in cases of conflict.
        combine_attrs: (optional), default = "drop_conflicts"
            Governs how to combine conflicting attrs. Is passed on to the xr merge
            functions.

        Returns
        -------
            xr.Dataset: Dataset with data from da_merge merged into the calling object
        """
        if self._ds.pr.has_processing_info():
            raise NotImplementedError(
                "Dataset contains processing information, this is not supported yet. "
                "Use ds.pr.remove_processing_info()."
            )

        ds_start = self._ds

        with contextlib.suppress(xr.MergeError, ValueError):
            # if there are no conflicts just merge using xr.merge
            return xr.merge(
                [ds_start, ds_merge],
                compat="no_conflicts",
                join="outer",
                combine_attrs=combine_attrs,
            )
        # merge by hand
        ensure_compatible_coords_dims(ds_merge, ds_start)

        vars_start = set(ds_start.data_vars)
        vars_merge = set(ds_merge.data_vars)
        vars_common = vars_start & vars_merge
        vars_only_start = vars_start - vars_common
        vars_only_merge = vars_merge - vars_common

        # merge variables which are only in one dataset and therefore trivially
        # mergeable
        ds_result = xr.merge(
            [ds_start[vars_only_start], ds_merge[vars_only_merge]],
            combine_attrs=combine_attrs,
        )

        # merge potentially problematic variables which are in both datasets
        for var in vars_common:
            logger.debug(f"merging for {var}")
            ds_result_new = merge_with_tolerance_core(
                da_start=ds_start[var],
                da_merge=ds_merge[var],
                tolerance=tolerance,
                error_on_discrepancy=error_on_discrepancy,
            )
            ds_result = xr.merge([ds_result, ds_result_new], combine_attrs="override")
        return ds_result



import itertools
import re

from loguru import logger

# basic units
_basic_units = ["g", "t"]

# prefixes
_si_unit_multipliers = ["", "k", "M", "G", "T", "P", "E", "Z", "Y"]

# combines basic units with prefixes
_units_prefixes = list(itertools.product(_si_unit_multipliers, _basic_units))
_units_prefixes = [i[0] + i[1] for i in _units_prefixes]

# build regexp to match the basic units with prefixes in units
_units_prefixes_regexp = "(" + "|".join(_units_prefixes) + ")"


def convert_unit_to_primap2(unit: str, entity: str) -> str:
    """Convert PRIMAP1 emissions module style units and units in similar formats to
    primap2 units.

    This function converts the emissions module style units which usually neither carry
    information about the substance nor about the time to primap2 units. The function
    also handles the exception cases where input units do contain information about the
    substance (e.g. GtC).

    Parameters
    ----------
    unit : str
        unit to convert
    entity : str
        entity for which the conversion takes place

    Returns
    -------
    unit : str
        converted unit
    """

    # check inputs
    if entity == "":
        logger.warning("Input entity is empty. Nothing converted.")
        return "error_" + unit + "_" + entity
    if unit == "":
        logger.warning("Input unit is empty. Nothing converted.")
        return "error_" + unit + "_" + entity

    # define exceptions
    exception_units = {
        "CO2eq": "CO2",  # convert to just CO2
        "CO2e": "CO2",  # convert to just CO2 (not for PRIMAP but e.g. NIRs)
        "CO₂eq": "CO2",  # convert to just CO2 (not for PRIMAP but e.g. NIRs)
        "<entity>N": "N",
        "C": "C",  # don't add variable here
    }

    # time information to add
    time_frame_str = " / yr"

    # remove spaces for more flexibility in input units
    unit = unit.replace(" ", "")

    # check if entity contains GWP information. If so discard
    # not needed for PRIMAP1 entities but when using the function for data reading
    entity_match = re.match(r"^[^\(\)\s]*", entity)
    entity = entity_match[0]

    # add entity and time frame to unit
    # special units will be replaced later
    unit_entity = unit + " " + entity + time_frame_str

    # check if unit has prefix
    match_pref = re.search(_units_prefixes_regexp, unit_entity)
    if match_pref is None:
        logger.warning("No unit prefix matched for unit. " + unit_entity)
        return "error_" + unit + "_" + entity

    # check if exception unit
    is_ex_unit = [
        re.match(
            _units_prefixes_regexp + ex_unit.replace("<entity>", entity) + "$", unit
        )
        is not None
        for ex_unit in exception_units
    ]

    if any(is_ex_unit):
        # we have an exception unit
        exception_unit = list(exception_units.keys())[is_ex_unit.index(True)]
        # first get the prefix and basic unit (e.g. Gt)
        pref_basic = match_pref.group(0)
        # now build the replacement
        converted_unit = (
            pref_basic + " " + exception_units[exception_unit] + time_frame_str
        )
    else:
        # standard unit
        converted_unit = unit_entity

    return converted_unit


def code_invalid_warn(code: str, message: str) -> str:
    """Log a warning and return an error code."""
    logger.warning(
        f"Category code {code!r} does not conform to specifications: {message}"
    )
    return "error_" + code


def convert_ipcc_code_primap_to_primap2(code: str) -> str:
    """Convert IPCC emissions category codes from PRIMAP1 emissions module style to
    primap2 style.

    Codes that are not part of the official hierarchy (starting with IPCM or CATM)
    are not converted but returned without the 'CAT' or 'IPC' prefix unless the
    conversion is explicitly defined in the code_mapping dict.
    TODO: add other primap terminologies?

    Parameters
    ----------
    code: str
        String containing the IPCC code in PRIMAP format (IPCC1996 and IPCC2006 can be
        converted). The PRIMAP format codes consist of upper case letters and numbers
        only. These are converted back to the original formatting which includes lower
        case letters and roman numerals.

    Returns
    -------
    code: str
        the category code in primap2 format

    Examples
    --------
    >>> convert_ipcc_code_primap_to_primap2("IPC1A3B34")
    '1.A.3.b.iii.4'
    """

    arabic_to_roman = {
        "1": "i",
        "2": "ii",
        "3": "iii",
        "4": "iv",
        "5": "v",
        "6": "vi",
        "7": "vii",
        "8": "viii",
        "9": "ix",
    }

    code_mapping = {
        "MAG": "M.AG",
        "MAGELV": "M.AG.ELV",
        "MBK": "M.BK",
        "MBKA": "M.BK.A",
        "MBKM": "M.BK.M",
        "MLULUCF": "M.LULUCF",
        "MMULTIOP": "M.MULTIOP",
        "M0EL": "M.0.EL",
        "MBIO": "M.BIO",
        "M3B4APF": "M.3.B.4.APF",
        "M3B4APD": "M.3.B.4.APD",
        "M3CAG": "M.3.C.AG",
        "M3C1AG": "M.3.C.1.AG",
        "M3C1AGSAV": "M.3.C.1.AG.SAV",
        "M3C1AGRES": "M.3.C.1.AG.RES",
        "M3D2LU": "M.3.D.2.LU",
        "M.AG": "M.AG",
        "M.AG.ELV": "M.AG.ELV",
        "M.BK": "M.BK",
        "M.BK.A": "M.BK.A",
        "M.BK.M": "M.BK.M",
        "M.LULUCF": "M.LULUCF",
        "M.MULTIOP": "M.MULTIOP",
        "M.0.EL": "M.0.EL",
        "M.BIO": "M.BIO",
    }

    if code[0:3] not in ["IPC", "CAT"]:
        # prefix = ""
        pure_code = code
    elif len(code) < 4:
        return code_invalid_warn(
            code, "Too short to be a PRIMAP IPCC code after " + "removal of prefix."
        )
    else:
        # prefix = code[0:3]
        pure_code = code[3:]

    # it's an IPCC code. convert it
    # check if it's a custom code (beginning with 'M'). Those have to be either mapped
    # explicitly using "code_mapping" or follow the normal structure after the "M"

    # check if a separator between prefix and code is used
    if len(pure_code) == 0:
        return code_invalid_warn(
            code, "Pure code has length 0. This should not be possible."
        )
    if pure_code[0] in [".", " ", "_", "-"]:
        pure_code = pure_code[1:]

    if pure_code[0] == "M":
        code_remaining = pure_code
        if pure_code in code_mapping.keys():
            new_code = code_mapping[pure_code]
            return new_code
        else:
            new_code = "M."
            code_remaining = code_remaining[1:]
            if len(code_remaining) == 0:
                return code_invalid_warn(
                    code, "Nothing follows the 'M' for an 'M'-code."
                )
    else:
        new_code = ""
        # only work with the part without 'IPC' or 'CAT'
        code_remaining = pure_code

    # if the code ends with a dot remove it
    if code_remaining[-1] == ".":
        code_remaining = code_remaining[:-1]

    # whenever we encounter dots as separators between levels, we ignore them.
    if code_remaining[0] == ".":
        code_remaining = code_remaining[1:]

    # actual conversion happening here
    # first level is a digit
    if code_remaining[0].isdigit():
        new_code = new_code + code_remaining[0]
    else:
        return code_invalid_warn(code, "No digit found on first level.")

    # second level is a letter
    if len(code_remaining) > 1:
        code_remaining = code_remaining[1:]
        if code_remaining[0] == ".":
            code_remaining = code_remaining[1:]
        # no need to check if code_remaining is emprty as we ensured that
        # the last char is not a dot (same in the following steps)
        if code_remaining[0].isalpha():
            new_code = new_code + "." + code_remaining[0]
        else:
            return code_invalid_warn(code, "No letter found on second level.")

        # third level is a number. might be more than one char, so use regexp
        if len(code_remaining) > 1:
            code_remaining = code_remaining[1:]
            if code_remaining[0] == ".":
                code_remaining = code_remaining[1:]
            match = re.match("^[0-9]+", code_remaining)
            if match is not None:
                new_code = new_code + "." + match.group(0)
            else:
                return code_invalid_warn(code, "No number found on third level.")

            # fourth level is a letter. has to be transformed to lower case
            if len(code_remaining) > len(match.group(0)):
                code_remaining = code_remaining[len(match.group(0)) :]
                if code_remaining[0] == ".":
                    code_remaining = code_remaining[1:]
                if code_remaining[0].isalpha():
                    new_code = new_code + "." + code_remaining[0].lower()
                else:
                    return code_invalid_warn(code, "No letter found on fourth level.")

                # fifth level is digit in PRIMAP1 format but roman numeral in IPCC
                # and PRIMAP2
                if len(code_remaining) > 1:
                    code_remaining = code_remaining[1:]
                    if code_remaining[0] == ".":
                        code_remaining = code_remaining[1:]
                    if code_remaining[0].isdigit():
                        new_code = new_code + "." + arabic_to_roman[code_remaining[0]]
                        len_level_5 = 1
                    else:
                        # try to match a roman numeral
                        match = re.match("^[ivx]{1,4}", code_remaining)
                        if match is not None:
                            new_code = new_code + "." + match.group(0)
                            len_level_5 = len(match.group(0))
                        else:
                            return code_invalid_warn(
                                code, "No digit or roman numeral found on fifth level."
                            )

                    # sixth and last level is a number.
                    if len(code_remaining) > len_level_5:
                        code_remaining = code_remaining[len_level_5:]
                        if code_remaining[0] == ".":
                            code_remaining = code_remaining[1:]
                        match = re.match("^[0-9]+", code_remaining)
                        if match is not None:
                            new_code = new_code + "." + match.group(0)
                            # check if anything left
                            if not code_remaining == match.group(0):
                                return code_invalid_warn(
                                    code, "Chars left after sixth level."
                                )
                        else:
                            return code_invalid_warn(
                                code, "No number found on sixth level."
                            )

    return new_code


def convert_entity_gwp_primap_to_primap2(entity_pm1: str) -> str:
    """Convert PRIMAP1 emissions module style entity names to primap2 style.

    The conversion only considers the GWP, currently the variable itself is
    unchanged.

    Currently the function uses a limited set of GWP values (defined in gwp_mapping) and
    works on a limited set of variables (defined in entities_gwp).

    Parameters
    entity: str
        entity to process

    Returns
    -------
    entity: str
        entity in PRIMAP2 format
    """

    entities_gwp = [
        "KYOTOGHG",
        "HFCS",
        "PFCS",
        "FGASES",
        "OTHERHFCS CO2EQ",
        "OTHERHFCS",
        "OTHERPFCS",
    ]

    # define the mapping of PRIMAP GWP specifications to PRIMAP2 GWP specification
    # no GWP given will be mapped to SAR
    gwp_mapping = {
        "SAR": "SARGWP100",
        "AR4": "AR4GWP100",
        "AR5": "AR5GWP100",
        "AR5CCF": "AR5CCFGWP100",  # not sure if implemented in scmdata units
        "AR6": "AR6GWP100",
    }

    # build regexp to match the GWP conversion variables
    regexp_str = "("
    for current_entity in entities_gwp:
        regexp_str = regexp_str + current_entity + "|"
    regexp_str = regexp_str[0:-1] + ")"

    # build regexp to match the GWPs
    regexp_str_gwps = "("
    for mapping in gwp_mapping.keys():
        regexp_str_gwps = regexp_str_gwps + mapping + "|"
    regexp_str_gwps = regexp_str_gwps[0:-1] + ")$"

    # check if entity in entities_gwp
    found = re.match(regexp_str, entity_pm1)
    if found is None:
        # not a basket entity which uses GWPs
        entity_pm2 = entity_pm1
    else:
        # check if GWP information present in entity
        match = re.search(regexp_str_gwps, entity_pm1)
        if match is None:
            # SAR GWPs are default in PRIMAP
            entity_pm2 = entity_pm1 + " (" + gwp_mapping["SAR"] + ")"
        else:
            gwp_out = match.group(0)
            # in this case the entity has to be replaced as well
            match = re.search(".*(?=" + gwp_out + ")", entity_pm1)
            if match.group() == "":
                logger.error(
                    "Confused: could not find entity which should be there."
                    " This indicates a bug in this function."
                )
                raise RuntimeError(
                    "Confused: could not find entity which should be there."
                    " This indicates a bug in this function."
                )
            else:
                entity_pm2 = match.group(0) + " (" + gwp_mapping[gwp_out] + ")"

    return entity_pm2



from collections.abc import Sequence

import numpy as np
import pandas as pd
import xarray as xr


def get_single_ts(
    *,
    time: pd.DatetimeIndex | None = None,
    data: np.ndarray | None = None,
    dims: Sequence[str] | None = None,
    coords: dict[str, str | Sequence[str]] | None = None,
    entity: str = "CH4",
    gwp_context: str | None = None,
) -> xr.DataArray:
    if time is None:
        time = pd.date_range("1850-01-01", "2022-01-01", freq="YS")
    if dims is None:
        dims = []
    if data is None:
        data = np.linspace(0.0, 1.0, len(time))
    if coords is None:
        coords = {}
    if gwp_context is None:
        name = entity
        attrs = {"entity": entity}
    else:
        name = f"{entity} ({gwp_context})"
        attrs = {"entity": entity, "gwp_context": gwp_context}
    return xr.DataArray(
        data,
        dims=["time", *dims],
        coords={"time": time, **coords},
        name=name,
        attrs=attrs,
    )



#!/usr/bin/env python
"""Tests for _units.py"""

import numpy as np
import pytest
import xarray as xr
import xarray.testing

from .utils import allclose, assert_equal


def test_roundtrip_quantify(opulent_ds: xr.Dataset):
    roundtrip = opulent_ds.pr.dequantify().pr.quantify()
    xarray.testing.assert_identical(roundtrip, opulent_ds)


def test_roundtrip_quantify_da(opulent_ds: xr.Dataset):
    da: xr.DataArray = opulent_ds["SF6 (SARGWP100)"]
    roundtrip = da.pr.dequantify().pr.quantify()
    assert_equal(roundtrip, da)


def test_convert_to_gwp(opulent_ds: xr.Dataset):
    da: xr.DataArray = opulent_ds["SF6"]
    da_converted = da.pr.convert_to_gwp("SARGWP100", "CO2 Gg / year")
    da_expected = opulent_ds["SF6 (SARGWP100)"]
    assert_equal(da_converted, da_expected)

    da_converted_like = da.pr.convert_to_gwp_like(da_expected)
    assert_equal(da_converted_like, da_expected)


def test_convert_to_gwp_like_missing(opulent_ds: xr.Dataset):
    da: xr.DataArray = opulent_ds["SF6"]
    da_gwp = da.pr.convert_to_gwp("SARGWP100", "CO2 Gg / year")

    del da_gwp.attrs["gwp_context"]
    with pytest.raises(ValueError, match="reference array has no gwp_context"):
        da.pr.convert_to_gwp_like(da_gwp)

    da_gwp = xr.full_like(da_gwp, np.nan)
    da_gwp.attrs["gwp_context"] = "SARGWP100"
    with pytest.raises(ValueError, match="reference array has no units attached"):
        da.pr.convert_to_gwp_like(da_gwp)


def test_convert_to_gwp_incompatible(opulent_ds: xr.Dataset):
    da: xr.DataArray = opulent_ds["SF6 (SARGWP100)"]
    with pytest.raises(ValueError, match="Incompatible gwp conversions"):
        da.pr.convert_to_gwp("AR5GWP", "CO2 Gg / year")


def test_convert_to_mass(opulent_ds: xr.Dataset):
    da: xr.DataArray = opulent_ds["SF6 (SARGWP100)"]
    da_converted = da.pr.convert_to_mass()
    da_expected = opulent_ds["SF6"]
    assert_equal(da_converted, da_expected)


def test_convert_round_trip(opulent_ds: xr.Dataset):
    da: xr.DataArray = opulent_ds["SF6"]
    assert da.attrs["entity"] == "SF6"
    da_gwp = da.pr.convert_to_gwp(gwp_context="AR4GWP100", units="Gg CO2 / year")
    da_rt = da_gwp.pr.convert_to_mass()
    assert_equal(da, da_rt)
    assert da_rt.attrs["entity"] == "SF6"
    assert isinstance(da_rt.attrs["entity"], str)


def test_convert_to_mass_missing_info(opulent_ds: xr.Dataset):
    da: xr.DataArray = opulent_ds["SF6"]
    with pytest.raises(
        ValueError,
        match="No gwp_context given and no gwp_context available in the attrs",
    ):
        da.pr.convert_to_mass()

    da = opulent_ds["SF6 (SARGWP100)"]
    del da.attrs["entity"]
    with pytest.raises(
        ValueError, match="No entity given and no entity available in the attrs"
    ):
        da.pr.convert_to_mass()


def test_context(opulent_ds: xr.Dataset):
    da: xr.DataArray = opulent_ds["SF6 (SARGWP100)"]
    with da.pr.gwp_context:
        da_converted = opulent_ds["SF6"].pint.to(da.pint.units)
    assert allclose(da, da_converted)



"""Models for the composite source generator."""

import typing
from collections.abc import Hashable

import attrs
import xarray as xr
from attr import define

import primap2
from primap2._data_format import ProcessingStepDescription


def equal_or_in(a, b):
    """Check if a == b (b str) or a in b (otherwise)."""
    if isinstance(b, str):
        return a == b
    else:
        return a in b


def match_selector(
    *, selector: dict[Hashable, str | list[str]], ts: xr.DataArray
) -> bool:
    """Check if a timeseries matches the selector."""
    for k, v in selector.items():
        if k == "entity":
            if not equal_or_in(ts.attrs["entity"], v):
                return False
        elif k == "variable":
            if not equal_or_in(ts.name, v):
                return False
        elif not equal_or_in(ts.coords[k], v):
            return False
    return True


@define(frozen=True, kw_only=True)
class PriorityDefinition:
    """
    Defines source priorities for composing a full dataset or a single timeseries.

    Attributes
    ----------
    priority_dimensions
        List of dimensions from which source timeseries are selected. Each priority has
        to specify all priority dimensions.
    priorities
        List of priority selections. Higher priority selections come first. Each
        selection consists of a dict which maps dimension names to values. Each
        selection has to specify all priority_dimensions, but may specify additional
        dimensions (fixed dimensions) to limit the selection to specific cases.
        You can use primap2.Not as a value for fixed dimensions to easily specify
        negative selections in the priorities. This should generally only be used if you
        want to change the priority of a source in specific circumstances. If you want
        to avoid using a source completely in specific cases, use exclude_input
        instead.
    exclude_input
        Set of selections from the input dataset to exclude from processing. Each
        selection consists of a dict which maps dimension names to values. Each
        selection can specify any number of priority dimensions or additional dimensions
        (fixed dimensions) to limit the selection.
        All input timeseries which match any selection will be skipped in processing.
        Note that after the exclusions, there still must be at least one applicable
        source for each result timeseries, otherwise errors will be raised. If you want
        to exclude timeseries from the result without raising errors, use
        exclude_result instead.
    exclude_result
        Set of selections from the result to exclude from all processing. Each selection
        consists of a dict which maps dimension names to values. Each selection can
        specify any number of additional dimensions (fixed dimensions) to limit the
        selection. Because it excludes timeseries from the result not from the input,
        it may not contain priority_dimensions.
        All timeseries which match any selection will be excluded entirely from
        processing and will be all-NaN in the result.

    Examples
    --------
        [{"area (ISO3)": ["MEX", "COL"], "source": "A"}, {"source": "B"}]
        would select source "A" as highest-priority source and source "B" as
        lower-priority source for Columbia and Mexico, but source "B" as
        highest-priority (and only) source for all other
        countries.
    """

    priority_dimensions: list[Hashable]
    priorities: list[dict[Hashable, str | list[str] | primap2.Not]]
    exclude_input: list[dict[Hashable, str | list[str]]] = attrs.field(default=[])
    exclude_result: list[dict[Hashable, str | list[str]]] = attrs.field(default=[])

    def limit(self, dim: Hashable, value: str) -> "PriorityDefinition":
        """Remove one fixed dimension by limiting to a single value.

        You can't remove priority dimensions, only fixed dimensions.
        """
        new_priorities = []
        for sel in self.priorities:
            if dim not in sel:
                new_priorities.append(sel)
                continue

            match_value = sel[dim]
            # for each possible type of match_value, skip this if it *doesn't* match
            if isinstance(match_value, primap2.Not):
                not_value = match_value.value
                if equal_or_in(value, not_value):
                    continue
            elif isinstance(match_value, str):
                if match_value != value:
                    continue
            elif value not in match_value:
                continue
            # now we know it is matching, filter out the matching dim
            new_priorities.append({k: v for k, v in sel.items() if k != dim})

        return PriorityDefinition(
            priority_dimensions=self.priority_dimensions,
            priorities=new_priorities,
            exclude_result=self.exclude_result,
            exclude_input=self.exclude_input,
        )

    def excludes_result(self, ts: xr.DataArray) -> bool:
        """Check if a selected result timeseries is excluded from processing."""
        return any(
            match_selector(selector=exclude_selector, ts=ts)
            for exclude_selector in self.exclude_result
        )

    def excludes_input(self, ts: xr.DataArray) -> bool:
        """Check if a selected input timeseries is excluded from processing."""
        return any(
            match_selector(selector=exclude_selector, ts=ts)
            for exclude_selector in self.exclude_input
        )

    def check_dimensions(self):
        """Raise an error if priorities or exclusions use wrong dimensions."""
        for sel in self.priorities:
            for dim in self.priority_dimensions:
                if dim not in sel:
                    raise ValueError(
                        f"In priority={sel}: missing priority dimension={dim}"
                    )
                if not isinstance(sel[dim], str):
                    raise ValueError(
                        f"In priority={sel}: specified multiple values for priority "
                        f"dimension={dim}, values={sel[dim]}"
                    )
        for sel in self.exclude_result:
            for dim in self.priority_dimensions:
                if dim in sel:
                    raise ValueError(
                        f"In result exclusion={sel}: excluded priority dimension={dim}"
                    )


class StrategyUnableToProcess(Exception):
    """The filling strategy is unable to process the given timeseries, possibly due
    to missing data."""

    def __init__(self, reason: str):
        """Specify the reason why the filling strategy is unable to process the data."""
        self.reason = reason


class FillingStrategyModel(typing.Protocol, Hashable):
    """
    Fill missing data in a timeseries using another timeseries.

    You can implement custom filling strategies to use with ``compose`` as long as they
    follow this protocol and are Hashable. To follow the protocol you need to implement
    the ``fill`` method with exactly the parameters and return types as defined below
    and define the ``type`` attribute (see below).

    To ensure that your class is hashable, you have to make sure instances are
    immutable after initialization (this helps caching). The easiest way to ensure
    immutability is usually to use the decorator ``attrs.define`` from the
    ``attrs`` package with the ``frozen=True`` argument.

    Attributes
    ----------
    type
        Short human-readable identifier for your strategy. Avoid special characters
        and spaces.
    """

    type: str

    def fill(
        self,
        *,
        ts: xr.DataArray,
        fill_ts: xr.DataArray,
        fill_ts_repr: str,
    ) -> tuple[xr.DataArray, list[ProcessingStepDescription]]:
        """Fill gaps in ts using data from the fill_ts.

        Using two input timeseries, this builds a composite timeseries and a description
        of the processing steps done. The input timeseries must not be modified.

        Usually, you want to fill missing data (NaNs) in the first timeseries `ts` using
        data from the second timeseries `fill_ts`. However, you are not limited to this,
        you could also check data in `ts` for consistency with data in `fill_ts` and
        discard non-conforming data points so that the resulting timeseries has more
        missing data points.

        Parameters
        ----------
        ts
            Base timeseries. Missing data (NaNs) in this timeseries will be filled or
            other processing is done.
            This function must not modify the data in ts, work on a copy instead.
        fill_ts
            Fill timeseries. Data from this timeseries will be used (possibly after
            modification) to fill missing data in the base timeseries or alter the
            base timeseries in a different form.
            This function must not modify the data in fill_ts.
        fill_ts_repr
            String representation of fill_ts. Human-readable short representation of
            the fill_ts (e.g. the source).

        Returns
        -------
            filled_ts, descriptions. filled_ts contains the result, where missing
            data in ts is (partly) filled using information from fill_ts or other
            processing is done using ts and fill_ts.
            descriptions contains human-readable, structured descriptions of how the
            data was processed, grouped by years for which the same processing steps
            were taken. Every year for which data in filled_ts is different from data
            in ts has to be described and no year for which data was not changed is
            allowed to be described.

        Raises
        ------
        StrategyUnableToProcess
            This exception is raised when the strategy is unable to process the given
            timeseries, possibly due to missing data (e.g. insufficient overlap of
            the two timeseries), bad numerical conditioning or other reasons.
            When this exception is raised, the strategy will be skipped and processing
            continues as if the strategy was not configured for this timeseries, i.e.
            the next applicable filling strategy is used. If no other applicable
            filling strategy is available, an error will be raised.
        """
        ...


@define(frozen=True)
class StrategyDefinition:
    """
    Defines filling strategies for a single timeseries.

    Attributes
    ----------
    strategies
        List of mappings from a timeseries selector to a filling strategy. When a
        timeseries will be used to fill missing data, the list will be checked from the
        start, and the first matching TimeseriesSelector determines the FillingStrategy.
        Example: [({"source": ["FAOSTAT", "UNFCCC]}, StraightStrategy()),
                  ({}, GlobalStrategy())]
        Note that the strategy can depend on fixed dimensions as well as priority
        dimensions.
        In practice, it is usually a good idea to include a default strategy using the
        empty selector {} which matches everything. It has to be the last entry -
        since it matches everything, all entries behind it will be ignored.

    """

    strategies: list[tuple[dict[Hashable, str | list[str]], FillingStrategyModel]]

    def find_strategy(self, fill_ts: xr.DataArray) -> FillingStrategyModel:
        """Find the strategy to use for the given filling timeseries."""
        try:
            return next(self.find_strategies(fill_ts))
        except StopIteration:
            raise KeyError(f"No matching strategy found for {fill_ts.coords}") from None

    def find_strategies(
        self, fill_ts: xr.DataArray
    ) -> typing.Generator[FillingStrategyModel, None, None]:
        """Yields all strategies to use for the timeseries, in configured order."""
        for selector, strategy in self.strategies:
            if match_selector(selector=selector, ts=fill_ts):
                yield strategy

    def limit(self, dim: Hashable, value: str) -> "StrategyDefinition":
        """Limit this strategy definition to strategies applicable with the limit."""
        return StrategyDefinition(
            strategies=[
                ({k: v for k, v in sel.items() if k != dim}, strat)
                for (sel, strat) in self.strategies
                if self.match_single_dim(selector=sel, dim=dim, value=value)
            ]
        )

    def check_dimensions(self, ds: xr.Dataset):
        """Raise an error if the strategy definition uses the wrong dimensions."""
        applicable_dimensions = set(ds.dims.keys()).union({"entity", "variable"})
        for sel, _ in self.strategies:
            for dim in sel:
                if dim not in applicable_dimensions:
                    raise ValueError(
                        f"In selector={sel!r}: {dim=} is not a valid dimension. Valid "
                        f"dimensions: {applicable_dimensions}."
                    )

    @staticmethod
    def match_single_dim(
        *, selector: dict[Hashable, str | list[str]], dim: Hashable, value: str
    ) -> bool:
        """Check if a literal value in one dimension can match the selector."""
        if dim in selector:
            return equal_or_in(value, selector[dim])
        else:
            return True



"""Tests for _alias_selection.py"""

import pytest
import xarray as xr
import xarray.testing

import primap2


@pytest.mark.parametrize(
    ["alias", "full_name"],
    [
        ("time", "time"),
        ("area", "area (ISO3)"),
        ("category", "category (IPCC 2006)"),
        ("cat", "category (IPCC 2006)"),
        ("animal", "animal (FAOSTAT)"),
        ("product", "product (FAOSTAT)"),
        ("scenario", "scenario (FAOSTAT)"),
        ("scen", "scenario (FAOSTAT)"),
        ("provenance", "provenance"),
        ("model", "model"),
        ("source", "source"),
        ("CO2", "CO2"),
        ("population", "population"),
    ],
)
def test_pr_getitem(opulent_ds, alias, full_name):
    da = opulent_ds.pr[alias]
    assert da.name == full_name


@pytest.mark.parametrize(
    ["alias", "full_name"],
    [
        ("time", "time"),
        ("area", "area (ISO3)"),
        ("category", "category (IPCC 2006)"),
        ("cat", "category (IPCC 2006)"),
        ("animal", "animal (FAOSTAT)"),
        ("product", "product (FAOSTAT)"),
        ("scenario", "scenario (FAOSTAT)"),
        ("scen", "scenario (FAOSTAT)"),
        ("provenance", "provenance"),
        ("model", "model"),
        ("source", "source"),
        ("CO2", "CO2"),
        ("population", "population"),
    ],
)
def test_pr_getitem_no_attrs(opulent_ds, alias, full_name):
    da = opulent_ds.notnull().pr[alias]
    assert da.name == full_name


@pytest.mark.parametrize(
    ["alias", "full_name"],
    [
        ("time", "time"),
        ("area", "area (ISO3)"),
        ("category", "category (IPCC 2006)"),
        ("cat", "category (IPCC 2006)"),
        ("animal", "animal (FAOSTAT)"),
        ("product", "product (FAOSTAT)"),
        ("scenario", "scenario (FAOSTAT)"),
        ("scen", "scenario (FAOSTAT)"),
        ("provenance", "provenance"),
        ("model", "model"),
        ("source", "source"),
    ],
)
def test_pr_alias_array(opulent_ds, alias, full_name):
    da = opulent_ds.pr["CO2"]
    actual = da.pr.sum(dim=alias)
    expected = da.sum(dim=full_name, keep_attrs=True)
    xr.testing.assert_identical(actual, expected)


def test_pr_loc_select(opulent_ds):
    sel_pr = opulent_ds.pr.loc[
        {
            "time": slice("2002", "2005"),
            "area": ["COL", "ARG"],
            "animal": "cow",
        }
    ]
    sel = opulent_ds.loc[
        {
            "time": slice("2002", "2005"),
            "area (ISO3)": ["COL", "ARG"],
            "animal (FAOSTAT)": "cow",
        }
    ]
    xr.testing.assert_identical(sel_pr, sel)


def test_pr_loc_select_da(opulent_ds):
    da = opulent_ds["CO2"]
    sel_pr = da.pr.loc[
        {
            "time": slice("2002", "2005"),
            "area": ["COL", "ARG"],
            "animal": "cow",
        }
    ]
    sel = da.loc[
        {
            "time": slice("2002", "2005"),
            "area (ISO3)": ["COL", "ARG"],
            "animal (FAOSTAT)": "cow",
        }
    ]
    xr.testing.assert_identical(sel_pr, sel)


def test_pr_loc_select_not(opulent_ds):
    sel_pr = opulent_ds.pr.loc[
        {
            "time": slice("2002", "2005"),
            "area": ["COL", "ARG"],
            "animal": primap2.Not("cow"),
            "category": primap2.Not(["0", "1"]),
        }
    ]
    sel = opulent_ds.loc[
        {
            "time": slice("2002", "2005"),
            "area (ISO3)": ["COL", "ARG"],
            "animal (FAOSTAT)": ["swine", "goat"],
            "category (IPCC 2006)": ["2", "3", "4", "5", "1.A", "1.B"],
        }
    ]
    xr.testing.assert_identical(sel_pr, sel)


def test_pr_loc_select_da_not(opulent_ds):
    da = opulent_ds["CO2"]
    sel_pr = da.pr.loc[
        {
            "time": slice("2002", "2005"),
            "area": ["COL", "ARG"],
            "animal": primap2.Not("cow"),
            "category": primap2.Not(["0", "1"]),
        }
    ]
    sel = da.loc[
        {
            "time": slice("2002", "2005"),
            "area (ISO3)": ["COL", "ARG"],
            "animal (FAOSTAT)": ["swine", "goat"],
            "category (IPCC 2006)": ["2", "3", "4", "5", "1.A", "1.B"],
        }
    ]
    xr.testing.assert_identical(sel_pr, sel)


def test_resolve_not(opulent_ds):
    result = primap2._selection.resolve_not(
        input_selector={
            "a": "1",
            "b": ["1", "2"],
            "animal (FAOSTAT)": primap2.Not("cow"),
            "area (ISO3)": primap2.Not(["MEX", "COL"]),
        },
        xarray_obj=opulent_ds,
    )
    assert len(result) == 4
    assert result["a"] == "1"
    assert result["b"] == ["1", "2"]
    assert len(result["animal (FAOSTAT)"]) == 2
    assert "swine" in result["animal (FAOSTAT)"]
    assert "goat" in result["animal (FAOSTAT)"]
    assert len(result["area (ISO3)"]) == 2
    assert "ARG" in result["area (ISO3)"]
    assert "BOL" in result["area (ISO3)"]


def test_resolve_not_da(opulent_ds):
    result = primap2._selection.resolve_not(
        input_selector={
            "a": "1",
            "b": ["1", "2"],
            "animal (FAOSTAT)": primap2.Not("cow"),
            "area (ISO3)": primap2.Not(["MEX", "COL"]),
        },
        xarray_obj=opulent_ds["CO2"],
    )
    assert len(result) == 4
    assert result["a"] == "1"
    assert result["b"] == ["1", "2"]
    assert len(result["animal (FAOSTAT)"]) == 2
    assert "swine" in result["animal (FAOSTAT)"]
    assert "goat" in result["animal (FAOSTAT)"]
    assert len(result["area (ISO3)"]) == 2
    assert "ARG" in result["area (ISO3)"]
    assert "BOL" in result["area (ISO3)"]



import numpy as np
import pytest
import xarray as xr
import xarray.testing

import primap2.csg
from primap2.csg import StrategyUnableToProcess
from primap2.tests.csg.utils import get_single_ts


@pytest.mark.parametrize(
    "strategy",
    [
        primap2.csg.SubstitutionStrategy(),
        primap2.csg.GlobalLSStrategy(),
    ],
    ids=lambda x: x.type,
)
def test_strategies_conform(strategy):
    """Test if strategies conform to the protocol for strategies."""
    ts = get_single_ts(data=1.0)
    ts[0] = np.nan
    fill_ts = get_single_ts(data=2.0)
    initial_ts = ts.copy(deep=True)
    initial_fill_ts = fill_ts.copy(deep=True)

    res, res_desc = strategy.fill(ts=ts, fill_ts=fill_ts, fill_ts_repr="B")

    # ts and fill_ts must not be modified
    xr.testing.assert_identical(ts, initial_ts)
    xr.testing.assert_identical(fill_ts, initial_fill_ts)

    assert isinstance(res, xr.DataArray)
    assert isinstance(res_desc, list)
    assert all(isinstance(step, primap2.ProcessingStepDescription) for step in res_desc)


def test_substitution_strategy():
    ts = get_single_ts(data=1.0)
    ts[0] = np.nan
    fill_ts = get_single_ts(data=2.0)

    result_ts, result_descriptions = primap2.csg.SubstitutionStrategy().fill(
        ts=ts, fill_ts=fill_ts, fill_ts_repr="B"
    )
    assert result_ts[0] == 2.0
    assert (result_ts[1:] == 1.0).all()
    assert len(result_descriptions) == 1
    assert result_descriptions[0].time == np.array(["1850"], dtype=np.datetime64)
    assert (
        result_descriptions[0].description
        == "substituted with corresponding values from B"
    )
    assert "source" not in result_ts.coords.keys()


def test_globalLS_strategy():
    ts = get_single_ts(data=1.0)
    fill_ts = get_single_ts(data=2.0)

    # nothing to fill
    result_ts, result_descriptions = primap2.csg.GlobalLSStrategy(
        allow_shift=False
    ).fill(ts=ts, fill_ts=fill_ts, fill_ts_repr="B")
    xr.testing.assert_allclose(ts, result_ts)
    assert len(result_descriptions) == 1
    assert len(result_descriptions[0].time) == 0  # comparison of results fails
    assert result_descriptions[0].description == "no additional data in B"

    # allow_shift = False
    ts[0] = np.nan
    result_ts, result_descriptions = primap2.csg.GlobalLSStrategy(
        allow_shift=False
    ).fill(ts=ts, fill_ts=fill_ts, fill_ts_repr="B")
    xr.testing.assert_allclose(get_single_ts(data=1.0), result_ts)
    assert len(result_descriptions) == 1
    assert result_descriptions[0].time == np.array(["1850"], dtype=np.datetime64)
    assert (
        result_descriptions[0].description
        == "filled with least squares matched data from B. Factor=0.500"
    )

    # allow_shift = True
    ts[1:5] = 0.5
    fill_ts[0:5] = 1.5
    result_ts, result_descriptions = primap2.csg.GlobalLSStrategy(
        allow_shift=True
    ).fill(ts=ts, fill_ts=fill_ts, fill_ts_repr="B")

    expected_ts = ts.copy(deep=True)
    expected_ts[0] = 0.5
    xr.testing.assert_allclose(expected_ts, result_ts)
    assert len(result_descriptions) == 1
    assert result_descriptions[0].time == np.array(["1850"], dtype=np.datetime64)
    assert (
        result_descriptions[0].description
        == "filled with least squares matched data from B. "
        "a*x+b with a=1.000, b=-1.000"
    )

    # error for negative emissions
    ts[1:5] = 0.5
    fill_ts[1:5] = 1.5
    fill_ts[0] = 0.1
    with pytest.raises(StrategyUnableToProcess):
        result_ts, result_descriptions = primap2.csg.GlobalLSStrategy(
            allow_shift=True, allow_negative=False
        ).fill(ts=ts, fill_ts=fill_ts, fill_ts_repr="B")

    # error for no overlap
    ts[1:5] = np.nan
    fill_ts[5:] = np.nan
    with pytest.raises(StrategyUnableToProcess):
        result_ts, result_descriptions = primap2.csg.GlobalLSStrategy(
            allow_shift=True, allow_negative=False
        ).fill(ts=ts, fill_ts=fill_ts, fill_ts_repr="B")

    # general
    assert "source" not in result_ts.coords.keys()



"""This file contains functions for reading of country GHG inventories
from National Inventory Reports (NIR), biannual Update Reports (BUR),
and other official country emissions inventories
Most of the functions in this file are exposed to the outside yet they
currently do not undergo the strict testing applied to the rest of PRIMAP2 as
they are added during the process of reading an preparing data for the PRIMAP-hist
update. Testing will be added in the future."""

import re

import pandas as pd


def nir_add_unit_information(
    df_nir: pd.DataFrame,
    *,
    unit_row: str | int,
    entity_row: str | int | None = None,
    regexp_entity: str,
    regexp_unit: str | None = None,
    manual_repl_unit: dict[str, str] | None = None,
    manual_repl_entity: dict[str, str] | None = None,
    default_unit: str,
) -> pd.DataFrame:
    """Add unit information to a National Inventory Report (NIR) style DataFrame.

    Add unit information to the header of an "entity-wide" file as
    present in the standard table format of National Inventory Reports (NIRs). The
    unit and entity information is extracted from combined unit and entity information
    in the row defined by `unit_row`. The parameters `regexp_unit` and `regexp_entity`
    determines how this is done by regular expressions for unit and entity.
    Additionally, manual mappings can be defined in the `manual_repl_unit` and
    `manual_repl_entity` dicts. For each column the routine tries to extract a unit
    using the regular expression. If this fails it looks in the `manual_repl_unit`
    dict for unit and in `manual_repl_entity` for entity information. If there is no
    information the default unit given in `default_unit` is used. In this case the
    analyzed value is used as entity unchanged.

    Parameters
    ----------
    df_nir : pd.DataFrame
        Pandas DataFrame with the table to process
    unit_row : str or int
        String "header" to indicate that the column header should be used to derive the
        unit information or an integer specifying the row to use for unit information.
        If entity and unit information are given in the same row use only unit_row.
    entity_row : str or int
        String "header" to indicate that the column header should be used to derive the
        unit information or an integer specifying the row to use for entity information.
        If entity and unit information are given in the same row use only unit_row
    regexp_entity : str
        regular expression that extracts the entity from the cell value
    regexp_unit : str (optional)
        regular expression that extracts the unit from the cell value
    manual_repl_unit : dict (optional)
        dict defining unit for given cell values
    manual_repl_entity : dict (optional)
        dict defining entity for given cell values
    default_unit : str
        unit to be used if no unit can be extracted an no unit is given

    Returns
    -------
    pd.DataFrame
        DataFrame with explicit unit information (as column header)
    """

    if manual_repl_unit is None:
        manual_repl_unit = {}

    if manual_repl_entity is None:
        manual_repl_entity = {}

    cols_to_drop = []

    # get the data to extract the units and entities from
    # can be either the header row or a regular row
    if unit_row == "header":
        values_for_units = list(df_nir.columns)
    else:
        # unit_row must be an integer
        values_for_units = list(df_nir.iloc[unit_row])
        cols_to_drop.append(unit_row)

    if entity_row is not None:
        if entity_row == "header":
            values_for_entities = list(df_nir.columns)
        else:
            values_for_entities = list(df_nir.iloc[entity_row])
            if entity_row != unit_row:
                cols_to_drop.append(entity_row)
    else:
        values_for_entities = values_for_units

    if regexp_unit is not None:
        re_unit = re.compile(regexp_unit)
    re_entity = re.compile(regexp_entity)

    units = values_for_units.copy()
    entities = values_for_entities.copy()

    for idx, value in enumerate(values_for_units):
        if str(value) in manual_repl_unit:
            units[idx] = manual_repl_unit[str(value)]
        else:
            if regexp_unit is not None:
                unit = re_unit.findall(str(value))
            else:
                unit = False

            if unit:
                units[idx] = unit[0]
            else:
                units[idx] = default_unit

    for idx, value in enumerate(values_for_entities):
        if str(value) in manual_repl_entity:
            entities[idx] = manual_repl_entity[str(value)]
        else:
            entity = re_entity.findall(str(value))
            if entity:
                entities[idx] = entity[0]
            else:
                entities[idx] = value

    new_header = [entities, units]

    df_out = df_nir.copy()
    df_out.columns = new_header
    if cols_to_drop:
        df_out = df_out.drop(df_out.index[cols_to_drop])

    return df_out


def nir_convert_df_to_long(
    df_nir: pd.DataFrame, year: int, header_long: list[str] | None = None
) -> pd.DataFrame:
    """Convert an entity-wide NIR table for a single year to a long format
    DataFrame.

    The input DataFrame is required to have the following structure:
    * Columns for category, original category name, and data in this order, where
    category and original category name form a multiindex.
    * Column header as multiindex for entity and unit
    A column for the year is added during the conversion.

    Parameters
    ----------
    df_nir: pd.DataFrame
        Pandas DataFrame with the NIR table to be converted
    year: int
        Year of the given data
    header_long: list, optional
        specify a non-standard column header, e.g. with only category code
        or orig_cat_name

    Returns
    -------
    pd.DataFrame
        converted DataFrame
    """
    if header_long is None:
        header_long = ["category", "orig_cat_name", "entity", "unit", "time", "data"]

    df_stacked = df_nir.stack([0, 1], dropna=True).to_frame()
    df_stacked.insert(0, "year", str(year))
    df_stacked = df_stacked.reset_index()
    df_stacked.columns = header_long
    return df_stacked



import contextlib
import datetime
import pathlib
import typing
from collections.abc import Hashable, Iterable, Mapping
from typing import IO

import msgpack
import numpy as np
import pandas as pd
import pint
import xarray as xr
from attr import define
from loguru import logger

from primap2._selection import translations_from_dims

from . import _accessor_base, pm2io
from ._units import ureg


def open_dataset(
    filename_or_obj: str | pathlib.Path | IO,
    group: str | None = None,
    chunks: int | dict | None = None,
    cache: bool | None = None,
    drop_variables: str | Iterable | None = None,
    backend_kwargs: dict | None = None,
) -> xr.Dataset:
    """Open and decode a dataset from a file or file-like object.

    Parameters
    ----------
    filename_or_obj : str, Path, file-like, or DataStore
        Strings and Path objects are interpreted as a path to a netCDF file
        or an OpenDAP URL and opened with h5py. Byte-strings or file-like
        objects are also supported.
    group : str, optional
        Path to the netCDF4 group in the given file to open.
    chunks : int or dict, optional
        If chunks is provided, it is used to load the new dataset into dask
        arrays. ``chunks={}`` loads the dataset with dask using a single
        chunk for all arrays.
    cache : bool, optional
        If True, cache data loaded from the underlying datastore in memory as
        NumPy arrays when accessed to avoid reading from the underlying data-
        store multiple times. Defaults to True unless you specify the `chunks`
        argument to use dask, in which case it defaults to False. Does not
        change the behavior of coordinates corresponding to dimensions, which
        always load their data from disk into a ``pandas.Index``.
    drop_variables: str or iterable, optional
        A variable or list of variables to exclude from being parsed from the
        dataset. This may be useful to drop variables with problems or
        inconsistent values.
    backend_kwargs: dict, optional
        A dictionary of keyword arguments to pass on to the backend. This
        may be useful when backend options would improve performance or
        allow user control of dataset processing.

    Returns
    -------
    dataset : Dataset
        The newly created dataset.

    Notes
    -----
    ``open_dataset`` opens the file with read-only access. When you modify
    values of a Dataset, even one linked to files on disk, only the in-memory
    copy you are manipulating in xarray is modified: the original file on disk
    is never touched.
    """
    ds = xr.open_dataset(
        filename_or_obj=filename_or_obj,
        group=group,
        chunks=chunks,
        cache=cache,
        drop_variables=drop_variables,
        backend_kwargs=backend_kwargs,
        engine="h5netcdf",
    ).pint.quantify(unit_registry=ureg)
    if "sec_cats" in ds.attrs:
        ds.attrs["sec_cats"] = list(ds.attrs["sec_cats"])
    if "publication_date" in ds.attrs:
        ds.attrs["publication_date"] = datetime.date.fromisoformat(
            ds.attrs["publication_date"]
        )
    for entity in ds:
        if entity.startswith("Processing of "):
            ds[entity].data = np.vectorize(
                lambda x: TimeseriesProcessingDescription.deserialize(x)
            )(ds[entity].data)
    return ds


class DatasetDataFormatAccessor(_accessor_base.BaseDatasetAccessor):
    """MixIn class which provides functions for checking the data format and saving
    of Datasets."""

    def ensure_valid(self) -> None:
        """Ensure this is a valid primap2 data set.

        Logs any deviations or non-standard properties. If the dataset violates any
        hard requirements of primap2 data sets, an exception is raised, otherwise the
        function simply returns.
        """
        if not isinstance(self._ds, xr.Dataset):
            logger.error("object is not an xarray Dataset.")
            raise ValueError("ds is not an xr.Dataset")

        ensure_valid_dimensions(self._ds)
        ensure_no_dimension_without_coordinates(self._ds)
        ensure_valid_coordinates(self._ds)
        ensure_valid_coordinate_values(self._ds)
        ensure_valid_data_variables(self._ds)
        ensure_valid_attributes(self._ds)

    def to_interchange_format(self, time_format: str = "%Y") -> pd.DataFrame:
        """Convert dataset to the interchange format.

        The interchange format consists of a pandas DataFrame in wide format with the
        meta data dictionary stored in the DataFrame's attrs dict.

        Parameters
        ----------
        time_format: str, optional
          string for strftime formatting of the "time" axis.

        Returns
        -------
        df: pd.DataFrame
        """
        dsd = self._ds.pr.dequantify()

        dsd["time"] = dsd["time"].dt.strftime(time_format)
        time_cols = list(dsd["time"].values)

        if "entity_terminology" in self._ds.attrs:
            entity_col = f"entity ({self._ds.attrs['entity_terminology']})"
        else:
            entity_col = "entity"

        dfs = []
        entities = []
        for x in dsd:
            if isinstance(x, str) and x.startswith("Processing of "):
                continue
            entities.append(x)
            df = (
                dsd[x]
                .to_dataset("time")
                .to_dataframe()
                .dropna(
                    how="all",
                    subset=time_cols,
                )
                .reset_index()
            )
            df[entity_col] = x
            df["unit"] = dsd[x].attrs.get("units", "no unit")
            dfs.append(df)

        df = pd.concat(dfs, ignore_index=True)

        add_coords = list(set(self._ds.coords) - set(self._ds.dims))
        df, dims = pm2io._data_reading.sort_columns_and_rows(
            df,
            dimensions=[dim for dim in dsd.dims if dim != "time"]
            + [entity_col, "unit"]
            + add_coords,
        )

        dimensions = {}
        all_dimensions = [str(dim) for dim in dsd.dims] + [entity_col, "unit"]
        for entity in entities:
            dims = [str(dim) for dim in dsd[entity].dims] + [entity_col, "unit"]
            if set(dims) == set(all_dimensions):
                # use default option for entities which use all the dataset's dimensions
                if "*" not in dimensions:
                    dimensions["*"] = all_dimensions
            else:
                dimensions[entity] = dims

        # all non-float dtypes are specified
        dtypes = {
            entity: str(dsd[entity].dtype)
            for entity in entities
            if dsd[entity].dtype != float
        }

        # add attrs for additional coords
        additional_coordinates = {}
        for coord in add_coords:
            coords_current = set(self._ds.coords[coord].coords) - {coord}
            if len(coords_current) != 1:
                logger.error(
                    f"Additional coordinate {coord!r} has more than one dimension, "
                    f"which is not supported."
                )
                raise ValueError(
                    f"Additional coordinate {coord!r} has more than one dimension"
                )

            additional_coordinates[coord] = next(iter(coords_current))

        df.attrs = {
            "attrs": self._ds.attrs,
            "time_format": time_format,
            "dimensions": dimensions,
        }
        if additional_coordinates:
            df.attrs["additional_coordinates"] = additional_coordinates
        if dtypes:
            df.attrs["dtypes"] = dtypes

        return df

    def to_netcdf(
        self,
        path: pathlib.Path | str,
        mode: str = "w",
        group: str | None = None,
        encoding: Mapping | None = None,
    ) -> bytes | None:
        """Write dataset contents to a netCDF file.

        Parameters
        ----------
        path : str or Path
            File path to which to save this dataset.
        mode : {"w", "a"}, default: "w"
            Write ('w') or append ('a') mode. If mode='w', any existing file at
            this location will be overwritten. If mode='a', existing variables
            will be overwritten.
        group : str, optional
            Path to the netCDF4 group in the given file to open. The group(s)
            will be created if necessary.
        encoding : dict, optional
            Nested dictionary with variable names as keys and dictionaries of
            variable specific encodings as values, e.g.,
            ``{"my_variable": {"dtype": "int16", "scale_factor": 0.1,
            "zlib": True}, ...}``

            This supports both the NetCDF4-style compression
            encoding parameters ``{"zlib": True, "complevel": 9}`` and the h5py
            ones ``{"compression": "gzip", "compression_opts": 9}``.
            This allows using any compression plugin installed in the HDF5
            library, e.g. LZF.
        """
        ds = self._ds.pint.dequantify()
        if "publication_date" in ds.attrs:
            ds.attrs["publication_date"] = ds.attrs["publication_date"].isoformat()
        for entity in ds:
            if (
                isinstance(entity, str)
                and entity.startswith("Processing of ")
                and ds[entity].data.dtype == object
            ):
                ds[entity].data = np.vectorize(lambda x: x.serialize())(ds[entity].data)
        return ds.to_netcdf(
            path=path,
            mode=mode,
            group=group,
            encoding=encoding,
            engine="h5netcdf",
            format="NETCDF4",
        )

    def remove_processing_info(self) -> xr.Dataset:
        """Return dataset with all variables with processing information removed."""
        return self._ds.drop_vars(
            [
                var
                for var in self._ds
                if isinstance(var, str) and var.startswith("Processing of ")
            ]
        )

    def has_processing_info(self) -> bool:
        """True if the dataset has processing information for at least one entity."""
        return any(
            isinstance(var, str) and var.startswith("Processing of ")
            for var in self._ds
        )

    def expand_dims(
        self,
        dim: str,
        coord_value: str,
        terminology: str | None = None,
    ) -> xr.Dataset:
        """
        Add a dimension and coordinate to dataset and if a terminology is given also to attrs.
        Dimension has length 1. primap2 equivalent of xarray's `expand_dims`

        To add a dimension to a PRIMAP2 DataArray use the native xarray method
        `expand_dims`. As PRIMAP2 DataArrays don't have the coordinate information
        in the attributes, no additional code is needed.

        Parameters
        ----------
        dim
            dimension to add. A coordinate with the same name will be added as well
        coord_value
            value for the
        terminology

        Returns
        -------
            dataset with added dimension and coordinate
        """

        ds_out = self._ds.copy(
            deep=True
        )  # deep=True necessary to keep attrs of original ds
        if terminology is not None:
            # get the translation for the attrs (use the shortest key to get the right key for the attrs)
            dim_to_add = f"{dim} ({terminology})"
            translations = translations_from_dims([dim_to_add])
            shortest_key = min(list(translations.keys()), key=len)
            # store in attributes
            ds_out.attrs[shortest_key] = translations[shortest_key]
        else:
            dim_to_add = dim

        ds_out = ds_out.expand_dims(dim={dim_to_add: [coord_value]})
        return ds_out


def split_dim_name(dim_name: str) -> tuple[str, str]:
    """Split a dimension name composed of the dimension, and the category set in
    parentheses into its parts."""
    try:
        dim, category_set = dim_name.split("(", 1)
    except ValueError:
        logger.error(f"{dim_name!r} not in the format 'dim (category_set)'.")
        raise ValueError(
            f"{dim_name!r} not in the format 'dim (category_set)'"
        ) from None
    return dim[:-1], category_set[:-1]


def split_var_name(var_name: str) -> tuple[str, str]:
    """Split a variable name composed of the entity and the gwp_context in parentheses
    into its parts."""
    try:
        entity, gwp_context = var_name.split("(", maxsplit=1)
    except ValueError:
        raise ValueError(
            f"Tried to split variable name {var_name!r}, but it does not specify a "
            f"gwp_context in parentheses."
        ) from None
    return entity[:-1], gwp_context[:-1]


def ensure_no_dimension_without_coordinates(ds: xr.Dataset):
    for dim in ds.dims:
        if dim not in ds.coords:
            logger.error(f"No coord found for dimension {dim!r}.")
            raise ValueError(f"dim {dim!r} has no coord")


def ensure_valid_coordinates(ds: xr.Dataset):
    additional_coords = set(ds.coords) - set(ds.dims)
    for coord in ds.coords:
        if not isinstance(coord, str):
            # Note: can not be tested properly because xarray deals poorly with
            # coordinate names which are not sortable with strings so the test never
            # gets here anyway
            logger.error(
                f"Coordinate {coord!r} is of type {type(coord)}, but "
                f"only strings are allowed."
            )
            raise ValueError(f"Coord key {coord!r} is not a string")
        elif coord in additional_coords:
            if " " in coord:
                logger.error(
                    f"Additional coordinate name {coord!r} contains a space, "
                    f"replace it with an underscore."
                )
                raise ValueError(f"Coord key {coord!r} contains a space")


def ensure_valid_attributes(ds: xr.Dataset):
    with contextlib.suppress(KeyError):
        reference = ds.attrs["references"]
        if not reference.startswith("doi:"):
            logger.info(f"Reference information is not a DOI: {reference!r}")
    with contextlib.suppress(KeyError):
        contact = ds.attrs["contact"]
        if "@" not in contact:
            logger.info(f"Contact information is not an email address: {contact!r}.")
    with contextlib.suppress(KeyError):
        publication_date = ds.attrs["publication_date"]
        if not isinstance(publication_date, datetime.date):
            logger.error(
                f"Publication date is not a datetime.date object: {publication_date!r}."
            )
            raise ValueError("Publication date is not a date object.")
    valid_attr_keys = {
        "references",
        "rights",
        "contact",
        "title",
        "comment",
        "institution",
        "area",
        "cat",
        "sec_cats",
        "scen",
        "entity_terminology",
        "publication_date",
    }
    unknown_attr_keys = set(ds.attrs.keys()) - valid_attr_keys
    if unknown_attr_keys:
        logger.warning(f"Unknown metadata in attrs: {unknown_attr_keys!r}, typo?")


def ensure_valid_data_variables(ds: xr.Dataset):
    ds = ensure_units_exist(ds)

    for key in ds:
        da: xr.DataArray = ds[key]  # type: ignore
        ensure_entity_and_units_exist(key, da)
        ensure_entity_and_units_valid(key, da)
        ensure_variable_name(str(key), da)

        if "gwp_context" in da.attrs:
            ensure_gwp_context_valid(str(key), da)
        else:
            ensure_not_gwp(key, da)

        if "described_variable" in da.attrs or (
            isinstance(key, str) and key.startswith("Processing of ")
        ):
            ensure_processing_variable_name(str(key), da)


def ensure_processing_variable_name(name: str, da: xr.DataArray) -> None:
    if "described_variable" not in da.attrs:
        logger.error(
            f"variable name {name!r} implies metadata variable, but 'described_variable'"
            f" is not defined in attrs."
        )
        raise ValueError(f"'described_variable' attr missing for {name!r}")
    if name != f"Processing of {da.attrs['described_variable']}":
        logger.error(
            f"variable name {name!r} is inconsistent with described_variable "
            f"{da.attrs['described_variable']!r}"
        )
        raise ValueError(
            f"variable name {name!r} inconsistent with described_variable"
            f" {da.attrs['described_variable']!r}"
        )


def ensure_entity_and_units_exist(key: Hashable, da: xr.DataArray):
    if "entity" not in da.attrs:
        logger.error(f"{key!r} has no entity declared in attributes.")
        raise ValueError(f"entity missing for {key!r}")

    if da.pint.units is None and da.dtype == float:
        logger.error(f"{key!r} is numerical (float) data, but has no units.")
        raise ValueError(f"units missing for {key!r}")


def ensure_units_exist(ds: xr.Dataset) -> xr.Dataset:
    try:
        # ensure that units are attached
        return ds.pint.quantify(unit_registry=ureg)
    except ValueError:  # raised by pint when units attached and in attrs
        logger.error("'units' in variable attrs, but data is quantified already.")
        raise


def ensure_entity_and_units_valid(key: Hashable, da: xr.DataArray):
    entity = da.attrs["entity"]
    units = da.pint.units
    with contextlib.suppress(pint.UndefinedUnitError):
        # if the entity is a gas and it is not converted to a gwp, the dimensions
        # should be compatible with an emission rate
        unit_entity = ureg(entity)
        if "gwp_context" not in da.attrs and not units.is_compatible_with(
            unit_entity * ureg.Gg / ureg.year
        ):
            logger.warning(
                f"{key!r} has a unit of {units}, which is not "
                f"compatible with an emission rate."
            )


def ensure_gwp_context_valid(key: str, da: xr.DataArray):
    units = da.pint.units
    gwp_context = da.attrs["gwp_context"]

    if units.dimensionality != {"[carbon]": 1, "[mass]": 1, "[time]": -1}:
        logger.error(
            f"{key!r} is a global warming potential, but the dimension is not "
            f"[CO2 * mass / time]."
        )
        raise ValueError(f"{key} has wrong dimensionality for gwp_context.")

    try:
        with ureg.context(gwp_context):
            pass
    except KeyError:
        logger.error(f"gwp_context {gwp_context!r} for {key!r} is not valid.")
        raise ValueError(f"Invalid gwp_context {gwp_context!r} for {key!r}") from None

    if "(" not in key or not key.endswith(")"):
        logger.warning(f"{key!r} has a gwp_context in attrs, but not in its " f"name.")


def ensure_not_gwp(key: Hashable, da: xr.DataArray):
    if (
        da.pint.units is not None
        and da.pint.units.dimensionality == {"[carbon]": 1, "[mass]": 1, "[time]": -1}
        and da.attrs["entity"] != "CO2"
    ):
        logger.warning(
            f"{key!r} has the dimension [CO2 * mass / time], but is "
            f"not CO2. gwp_context missing?"
        )


def ensure_variable_name(key: str, da: xr.DataArray):
    if "gwp_context" in da.attrs:
        common_name = f"{da.attrs['entity']} ({da.attrs['gwp_context']})"
    else:
        common_name = da.attrs["entity"]

    if common_name != key:
        logger.info(f"The name {key!r} is not in standard format {common_name!r}.")


def ensure_valid_coordinate_values(ds: xr.Dataset):
    if "provenance" in ds.dims:
        wrong_values = set(ds["provenance"].data) - {"projected", "measured", "derived"}
        if wrong_values:
            logger.error(f"provenance contains invalid values: {wrong_values!r}")
            raise ValueError(f"Invalid provenance: {wrong_values!r}")


def ensure_valid_dimensions(ds: xr.Dataset):
    required_direct_dims = {"time", "source"}
    required_indirect_dims = {"area"}
    optional_direct_dims = {"provenance", "model"}
    optional_indirect_dims = {"cat", "scen"}  # sec_cats is special

    for req_dim in required_direct_dims:
        if req_dim not in ds.dims:
            logger.error(f"{req_dim!r} not found in dims, but is required.")
            raise ValueError(f"{req_dim!r} not in dims")

        for var in ds:
            if (
                isinstance(var, str)
                and var.startswith("Processing of ")
                and req_dim == "time"
            ):
                if req_dim in ds[var].dims:
                    logger.error(
                        f"{var!r} is a metadata variable, but 'time' is a dimension."
                    )
                    raise ValueError(
                        f"{var!r} contains metadata, but carries 'time' dimension"
                    )
            else:
                if req_dim not in ds[var].dims:
                    logger.error(
                        f"{req_dim!r} not found in dims for variable {var!r}, but is required."
                    )
                    raise ValueError(f"{req_dim!r} not in dims for variable {var!r}")

    required_indirect_dims_long = []
    for req_dim in required_indirect_dims:
        if req_dim not in ds.attrs:
            logger.error(
                f"{req_dim!r} not found in attrs, required dimension is therefore"
                f" undefined."
            )
            raise ValueError(f"{req_dim!r} not in attrs")

        if ds.attrs[req_dim] not in ds.dims:
            logger.error(
                f"{ds.attrs[req_dim]!r} defined as {req_dim!r} dimension, but not found"
                f" in dims."
            )
            raise ValueError(f"{req_dim!r} dimension not in dims")

        required_indirect_dims_long.append(ds.attrs[req_dim])

    included_optional_dims = []
    for opt_dim in optional_indirect_dims:
        if opt_dim in ds.attrs:
            long_name = ds.attrs[opt_dim]
            included_optional_dims.append(long_name)
            if long_name not in ds.dims:
                logger.error(
                    f"{long_name!r} defined as {opt_dim}, but not found in dims."
                )
                raise ValueError(f"{opt_dim!r} not in dims")

    if "sec_cats" in ds.attrs:
        for sec_cat in ds.attrs["sec_cats"]:
            included_optional_dims.append(sec_cat)
            if sec_cat not in ds.dims:
                logger.error(
                    f"Secondary category {sec_cat!r} defined, but not found in dims: "
                    f"{ds.dims}."
                )
                raise ValueError(f"Secondary category {sec_cat!r} not in dims")

    if "sec_cats" in ds.attrs and "cat" not in ds.attrs:
        logger.warning(
            "Secondary category defined, but no primary category defined, weird."
        )

    all_dims = set(ds.dims.keys())
    unknown_dims = (
        all_dims
        - required_direct_dims
        - set(required_indirect_dims_long)
        - optional_direct_dims
        - set(included_optional_dims)
    )

    if unknown_dims:
        logger.warning(
            f"Dimension(s) {unknown_dims} unknown, likely a typo or missing in"
            f" sec_cats."
        )

    for dim in required_indirect_dims.union(optional_indirect_dims):
        if dim in ds.attrs:
            split_dim_name(ds.attrs[dim])


@define(frozen=True, kw_only=True)
class ProcessingStepDescription:
    """Structured description of a processing step done on a timeseries."""

    time: np.ndarray[np.datetime64] | typing.Literal["all"]
    function: str
    description: str
    source: str | None = None

    def __str__(self) -> str:
        if self.source is None:
            return (
                f"Using function={self.function} for times={self.time}:"
                f" {self.description}"
            )
        else:
            return (
                f"Using function={self.function} with source={self.source} for "
                f"times={self.time}: {self.description}"
            )

    def unstructure(self) -> dict[str, typing.Any]:
        """Convert into basic python types."""
        return {
            "time": "all"
            if isinstance(self.time, str) and self.time == "all"
            else list(
                np.datetime_as_string(
                    self.time,
                    unit="Y",
                )
            ),
            "description": self.description,
            "function": self.function,
            "source": self.source,
        }

    @classmethod
    def structure(cls, u: dict[str, typing.Any]) -> "ProcessingStepDescription":
        """Initialize from basic python types as created by "unstructure"."""
        time = u.pop("time")
        return cls(
            time="all" if time == "all" else np.array(time, dtype=np.datetime64), **u
        )


@define(frozen=True)
class TimeseriesProcessingDescription:
    """Structured description of all processing steps done on a timeseries."""

    steps: list[ProcessingStepDescription]

    def __str__(self) -> str:
        return "\n".join(str(step) for step in self.steps)

    def serialize(self) -> bytes:
        """Convert into binary data, e.g. for saving to disk."""
        return msgpack.packb(
            {"steps": [x.unstructure() for x in self.steps]}, use_bin_type=True
        )

    @classmethod
    def deserialize(cls, b: bytes) -> "TimeseriesProcessingDescription":
        """Parse from binary data as produced by "serilize"."""
        ust = msgpack.unpackb(b, raw=False, use_list=False)
        return cls(steps=[ProcessingStepDescription.structure(x) for x in ust["steps"]])



import csv
import itertools
import re
from pathlib import Path

import numpy as np
import pandas as pd
import strictyaml as sy
import xarray as xr
from loguru import logger
from ruamel.yaml import YAML

# entity is mandatory in the interchange format because it is transformed
# into the variables
# unit is mandatory in the interchange format because it is transformed
# into the pint units
INTERCHANGE_FORMAT_MANDATORY_COLUMNS = ["area", "source", "entity", "unit"]
INTERCHANGE_FORMAT_OPTIONAL_COLUMNS = ["category", "scenario", "provenance", "model"]
INTERCHANGE_FORMAT_COLUMN_ORDER = [
    "source",
    "scenario",
    "provenance",
    "model",
    "area",
    "entity",
    "unit",
    "category",
    "orig_cat_name",
    "cat_name_translation",
]

INTERCHANGE_FORMAT_STRICTYAML_SCHEMA = sy.Map(
    {
        sy.Optional("data_file"): sy.Str(),
        "attrs": sy.Map(
            {
                "area": sy.Str(),
                sy.Optional("cat"): sy.Str(),
                sy.Optional("sec_cats"): sy.Seq(sy.Str()),
                sy.Optional("scen"): sy.Str(),
                sy.Optional("references"): sy.Str(),
                sy.Optional("rights"): sy.Str(),
                sy.Optional("contact"): sy.Str(),
                sy.Optional("title"): sy.Str(),
                sy.Optional("comment"): sy.Str(),
                sy.Optional("institution"): sy.Str(),
                sy.Optional("entity_terminology"): sy.Str(),
                sy.Optional("publication_date"): sy.Datetime(),
            }
        ),
        "dimensions": sy.MapPattern(sy.Str(), sy.Seq(sy.Str())),
        "time_format": sy.Str(),
        sy.Optional("additional_coordinates"): sy.MapPattern(sy.Str(), sy.Str()),
        sy.Optional("dtypes"): sy.MapPattern(sy.Str(), sy.Str()),
    }
)


def dates_to_dimension(ds: xr.Dataset, time_format: str = "%Y") -> xr.DataArray:
    """
    This function converts a xr.Dataset where each time point is an individual
    data variable to a xr.DataArray with a time dimension.
    All variables which are not in the index are assumed to be time points.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset with individual time points as data variables

    time_format : str
        format string for the time points. Default: %Y (for year only)

    Returns
    -------
    reduced: xr.DataArray
        xr.DataArray with the time as a dimension and time points as values
    """
    empty_vars = (x for x in ds if ds[x].count() == 0)
    da = ds.drop_vars(empty_vars).to_array("time")
    da["time"] = pd.to_datetime(da["time"].values, format=time_format, exact=False)
    return da


def metadata_for_variable(unit: str, variable: str) -> dict[str, str]:
    """Convert a primap2 unit and variable key to a metadata dict.

    Derives the information needed for the data variable's attrs dict from the unit
    and the variable's key.
    Takes GWP information from the variable name in primap2 style (if present).

    Parameters
    ----------
    unit: str
        unit to be stored in the attrs dict
    variable: str
        entity to process

    Returns
    -------
    attrs: dict
        attrs for use in DataArray.attrs

    Examples
    --------

    >>> metadata_for_variable("Gg CO2 / year", "Kyoto-GHG (SARGWP100)")
    {'units': 'Gg CO2 / year', 'gwp_context': 'SARGWP100', 'entity': 'Kyoto-GHG'}

    """
    attrs = {}

    if not isinstance(unit, str):
        unit = ""

    if unit != "no unit":
        attrs["units"] = unit

    regex_gwp = r"\s\(([A-Za-z0-9]*)\)$"
    gwp = re.findall(regex_gwp, variable)

    if gwp:
        attrs["gwp_context"] = gwp[0]

        regex_variable = r"^(.*)\s\([a-zA-z0-9]*\)$"
        entity = re.findall(regex_variable, variable)
        if not entity:
            logger.error(f"Can't extract entity from {variable}")
            raise ValueError(f"Can't extract entity from {variable}")
        attrs["entity"] = entity[0]
    else:
        attrs["entity"] = variable
    return attrs


def write_interchange_format(
    filepath: str | Path, data: pd.DataFrame, attrs: dict | None = None
) -> None:
    """Write dataset in interchange format to disk.

    Writes an interchange format dataset consisting of a pandas Dataframe and an
    additional meta data dict to disk. The data is stored in a csv file while the
    additional metadata is written to a yaml file.

    Parameters
    ----------
    filepath: str or pathlib.Path
        path and filename stem for the dataset. If a file ending is given it will be
        ignored and replaced by .csv for the data and .yaml for the metadata

    data: pandas.DataFrame
        DataFrame in PRIMAP2 interchange format

    attrs: dict, optional
        Interchange format meta data dict. Default: use data.attrs .
    """
    if attrs is None:
        attrs = data.attrs.copy()
    else:
        attrs = attrs.copy()

    # make sure filepath is a Path object
    filepath = Path(filepath)
    data_file = filepath.parent / (filepath.stem + ".csv")
    meta_file = filepath.parent / (filepath.stem + ".yaml")

    # write the data
    data.to_csv(data_file, index=False, quoting=csv.QUOTE_NONNUMERIC)

    attrs["data_file"] = data_file.name

    yaml = YAML()
    # settings for strictyaml compatibility: don't use flow style or aliases
    yaml.default_flow_style = False
    yaml.representer.ignore_aliases = lambda x: True
    with meta_file.open("w") as fd:
        yaml.dump(attrs, fd)


def read_interchange_format(
    filepath: str | Path,
) -> pd.DataFrame:
    """Read a dataset in the interchange format from disk into memory.

    Reads an interchange format dataset from disk. The data is stored
    in a csv file while the additional metadata is stored in a yaml
    file. This function takes the yaml file as parameter, the data file is specified
    in the yaml file. If no or a wrong ending is given the function tries to load
    a file by the same name with the ending `.yaml`.

    Parameters
    ----------
    filepath: str or pathlib.Path
        path and filename for the dataset (the yaml file, not data file).

    Returns
    -------
    data: pandas.DataFrame
        DataFrame with the read data in PRIMAP2 interchange format
    """
    filepath = Path(filepath)
    if not filepath.exists():
        filepath = filepath.with_suffix(".yaml")
    with filepath.open() as meta_file:
        yaml = sy.load(meta_file.read(), schema=INTERCHANGE_FORMAT_STRICTYAML_SCHEMA)
        meta_data = yaml.data

    data_file = filepath.parent / meta_data.get("data_file", filepath.stem + ".csv")
    if not data_file.exists():
        raise FileNotFoundError(f"Data file not found at {data_file}.")

    data = pd.read_csv(data_file, dtype=object)
    data.attrs = meta_data

    # strictyaml parses a datetime, we only want a date
    if "publication_date" in data.attrs["attrs"]:
        data.attrs["attrs"]["publication_date"] = data.attrs["attrs"][
            "publication_date"
        ].date()
    # already read in
    if "data_file" in data.attrs:
        del data.attrs["data_file"]

    return data


def from_interchange_format(
    data: pd.DataFrame,
    attrs: dict | None = None,
    max_array_size: int = 1024 * 1024 * 1024,
) -> xr.Dataset:
    """Convert dataset from the interchange format to the standard PRIMAP2 format.

    Converts an interchange format DataFrame with added metadata to a PRIMAP2 xarray
    data structure. All column names and attrs are expected to be already in PRIMAP2
    format as defined for the interchange format. The attrs dict is given explicitly as
    the attrs functionality in pandas is experimental.

    Parameters
    ----------
    data: pd.DataFrame
        pandas DataFrame in PRIMAP2 interchange format.
    attrs: dict, optional
        attrs dict as defined for the PRIMAP2 interchange format. Default: use
        data.attrs.
    max_array_size: int, optional
        Maximum permitted projected array size. Larger sizes will raise an exception.
        Default: 1 G, corresponding to about 4 GB of memory usage.

    Returns
    -------
    obj: xr.Dataset
        xr dataset with the converted data
    """
    if attrs is None:
        attrs = data.attrs.copy()

    if "entity_terminology" in attrs["attrs"]:
        entity_col = f"entity ({attrs['attrs']['entity_terminology']})"
    else:
        entity_col = "entity"

    if "additional_coordinates" not in attrs:
        attrs["additional_coordinates"] = {}

    # build dicts for additional coordinates
    add_coord_mapping_dicts = {}
    for coord in attrs["additional_coordinates"].keys():
        values = data[[coord, attrs["additional_coordinates"][coord]]]
        values = values.drop_duplicates()
        dim_values = list(values[attrs["additional_coordinates"][coord]])
        coord_values = list(values[coord])
        if len(coord_values) != len(set(dim_values)):
            logger.error(
                f"Different secondary coordinate values for given first coordinate "
                f"value for {coord}."
            )
            raise ValueError(
                f"Different secondary coordinate values for given first coordinate "
                f"value for {coord}."
            )

        add_coord_mapping_dicts[coord] = dict(
            zip(dim_values, coord_values, strict=False)
        )

    # drop additional coordinates. make a copy first to not alter input DF
    data_drop = data.drop(columns=attrs["additional_coordinates"].keys(), inplace=False)

    # find the time columns
    if_index_cols = set(itertools.chain(*attrs["dimensions"].values()))
    time_cols = set(data_drop.columns.values) - if_index_cols

    # convert to xarray
    data_xr = data_drop.to_xarray()
    index_cols = if_index_cols - {"unit", "time"}
    data_xr = data_xr.set_index({"index": list(index_cols)})
    # take the units out as they increase dimensionality and we have only one unit per
    # entity/variable
    units = data_xr["unit"]
    del data_xr["unit"]

    # build full dimensions dict from specification with default from entry "*"
    entities = np.unique(data_xr[entity_col].values)
    dimensions = attrs["dimensions"]
    for entity in entities:
        if entity not in dimensions:
            dimensions[entity] = dimensions["*"]
    if "*" in dimensions:
        del dimensions["*"]

    # build full dtypes dict from specification with default float
    dtypes = attrs.get("dtypes", {})
    for entity in entities:
        if entity not in dtypes:
            dtypes[entity] = "float"

    # check resulting shape to estimate memory consumption
    dim_lens = {dim: len(np.unique(data_xr[dim].dropna("index"))) for dim in index_cols}
    dim_lens["time"] = len(time_cols)
    shapes = []
    for _, dims in dimensions.items():
        shapes.append([dim_lens[dim] for dim in dims if dim != "unit"])
    array_size = sum(np.prod(shape) for shape in shapes)
    logger.debug(f"Expected array shapes: {shapes}, resulting in size {array_size:,}.")
    if array_size > max_array_size:
        logger.error(
            f"Set with {len(shapes)} entities and a total of {len(index_cols)} "
            f"dimensions will have a size of {array_size:,} "
            f"due to the shapes {shapes}. Aborting to avoid out-of-memory errors. To "
            f"continue, raise max_array_size (currently {max_array_size:,})."
        )
        raise ValueError(
            f"Resulting array too large: {array_size:,} > {max_array_size:,}. To "
            f"continue, raise max_array_size."
        )

    # convert time to dimension and entity to variables.
    if "time_format" in attrs:
        da = dates_to_dimension(data_xr, time_format=attrs["time_format"])
    else:
        da = dates_to_dimension(data_xr)
    data_vars = {}
    for entity, dims in dimensions.items():
        da_entity = da.loc[{entity_col: entity}]
        # we still have a full MultiIndex, so trim it to the relevant dimensions
        da_entity = da_entity.reset_index(list(index_cols - set(dims)), drop=True)
        # depending on the version of xarray, an atomic, 0-dimensional coord
        # for the entity remains. we have to remove it to be able to combine the
        # dataset afterwards.
        if entity_col in da_entity.coords:
            da_entity = da_entity.drop_vars(entity_col)
        # now we can safely unstack the index
        data_vars[entity] = da_entity.unstack("index").astype(dtypes[entity])

    data_xr = xr.Dataset(data_vars)

    # add the additional coordinates
    for coord in attrs["additional_coordinates"].keys():
        dim_values_xr = list(data_xr[attrs["additional_coordinates"][coord]].values)
        coord_values_ordered = [
            add_coord_mapping_dicts[coord][value] for value in dim_values_xr
        ]
        data_xr = data_xr.assign_coords(
            {
                coord: xr.DataArray(
                    data=np.array(coord_values_ordered),
                    coords={
                        attrs["additional_coordinates"][coord]: data_xr.coords[
                            attrs["additional_coordinates"][coord]
                        ]
                    },
                    dims=[attrs["additional_coordinates"][coord]],
                )
            }
        )

    # fill the entity/variable attributes
    for variable in data_xr:
        csv_units = np.unique(units.loc[{entity_col: variable}], equal_nan=True)
        if len(csv_units) > 1 and any(isinstance(x, str) for x in csv_units):
            logger.error(
                f"More than one unit for entity {variable!r}: {csv_units!r}. "
                + "There is an error in the unit harmonization."
            )
            raise ValueError(f"More than one unit for {variable!r}: {csv_units!r}.")
        data_xr[variable].attrs = metadata_for_variable(csv_units[0], variable)

    # add the dataset wide attributes
    data_xr.attrs = attrs["attrs"]

    data_xr = data_xr.pr.quantify()

    data_xr.pr.ensure_valid()
    return data_xr
